"""dataset commands for cmem command line interface."""
import json
import os
import re

import click

import requests.exceptions

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.commands.resource import resource
from cmem.cmemc.cli.context import ApplicationContext
from cmem.cmemc.cli.utils import struct_to_table
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.workspace.search import list_items
from cmem.cmempy.workspace.projects.project import (
    get_projects
)
from cmem.cmempy.workspace.projects.datasets.dataset import (
    create_dataset,
    delete_dataset,
    get_dataset
)
from cmem.cmempy.workspace.projects.resources.resource import (
    create_resource,
    get_resource_response,
    resource_exist
)
from cmem.cmempy.workspace import (
    get_task_plugin_description,
    get_task_plugins
)

DATASET_FILTER_TYPES = sorted(['project', 'regex', 'tag', 'type'])
DATASET_FILTER_HELP_TEXT = (
    "Filter datasets based on meta data. First parameter --filter"
    f" CHOICE can be one of {str(DATASET_FILTER_TYPES)}."
    " The second parameter is based on CHOICE."
)


def _get_dataset_tag_labels(dataset_) -> list:
    """Output a list of tag labels from a single dataset."""
    tag_labels = []
    for tag in dataset_["tags"]:
        tag_labels.append(tag["label"])
    return tag_labels


def _get_datasets_filtered(datasets, filter_name, filter_value):
    """Get dataset filtered according to filter name and value.

    Args:
        datasets: list of datasets to filter
        filter_name (str): one of DATASET_FILTER_TYPES
        filter_value (str|int): value according to filter

    Returns:
        list of filtered datasets from the get_query_status API call

    Raises:
        ValueError
    """
    if filter_name not in DATASET_FILTER_TYPES:
        raise ValueError(
            f"{filter_name} is an unknown filter name. "
            f"Use one of {DATASET_FILTER_TYPES}."
        )
    # filter by project ID
    if filter_name == "project":
        return [_ for _ in datasets
                if _["projectId"] == filter_value]
    # filter by regex on the label
    if filter_name == "regex":
        return [
            _ for _ in datasets
            if re.search(filter_value, _["label"])
        ]
    # filter by dataset type
    if filter_name == "type":
        return [
            _ for _ in datasets
            if re.search(filter_value, _["pluginId"])
        ]
    # filter by tag label
    if filter_name == "tag":
        return [
            _ for _ in datasets
            if filter_value in _get_dataset_tag_labels(_)
        ]
    # default is unfiltered
    return datasets


def _validate_and_split_dataset_id(dataset_id):
    """Validate and split cmemc dataset ID.

    Args:
        dataset_id (str): The cmemc dataset ID in the workspace.

    Raises:
        ValueError: in case the dataset ID is not splittable
    """
    try:
        project_part = dataset_id.split(":")[0]
        dataset_part = dataset_id.split(":")[1]
    except IndexError as error:
        raise ValueError(
            f"{dataset_id} is not a valid dataset ID. Use the "
            "'cmemc dataset list' command to get a list of existing datasets."
        ) from error
    return project_part, dataset_part


def _upload_file_resource(
        app=None,
        project_id=None,
        local_file_name=None,
        remote_file_name=None,
        replace=False
):
    """Upload a local file as a dataset resource to a project.

    Args:
        app (ApplicationContext): the click cli app context.
        project_id (str): The project ID in the workspace.
        local_file_name (str): The path to the local file name
        remote_file_name (str): The remote file name
        replace (bool): Replace resource if needed.

    Raises:
        ValueError: if resource exists and no replace
    """
    exist = resource_exist(
        project_name=project_id, resource_name=remote_file_name
    )
    if exist and not replace:
        raise ValueError(
            f"A file resource with the name '{remote_file_name}' already "
            "exists in this project. \n"
            "Please rename the file or use the '--replace' "
            "parameter in order to overwrite the remote file."
        )
    if exist:
        app.echo_info(
            f"Replace content of {remote_file_name} with content from "
            f"{local_file_name} in project {project_id} ... ",
            nl=False
        )
    else:
        app.echo_info(
            f"Upload {local_file_name} as a file resource "
            f"{remote_file_name} to project {project_id} ... ",
            nl=False
        )
    create_resource(
        project_name=project_id,
        resource_name=remote_file_name,
        file_resource=click.open_file(local_file_name, "rb"),
        replace=replace
    )
    app.echo_success("done")


def _get_metadata_out_of_parameter(parameter_dict):
    """Extract metadata keys out of the parameter dict.

    Args:
        parameter_dict (dict): the dictionary of given parameters.

    Returns:
        The dictionary of only the known metadata fields.
    """
    metadata_dict = {}
    if "label" in parameter_dict:
        metadata_dict["label"] = parameter_dict["label"]
    if "description" in parameter_dict:
        metadata_dict["description"] = parameter_dict["description"]
    return metadata_dict


def _extend_parameter_with_metadata(
        app,
        parameter_dict=None,
        dataset_type=None,
        dataset_file=None
):
    """Extend the parameter with label if needed.

    Args:
        app (ApplicationContext): the click cli app context.
        parameter_dict (dict): The dictionary of given dataset parameters
        dataset_type (str): The dataset type ID
        dataset_file (str): The path of the local file

    Returns:
        An extended parameter dictionary (label + file)
    """
    if "label" not in parameter_dict:
        label = f"Unnamed {dataset_type} dataset"
        if dataset_file:
            label = os.path.basename(dataset_file)
        if "file" in parameter_dict:
            label = parameter_dict["file"]
        app.echo_warning("Missing dataset label (-p label xxx) - "
                         f"this generated label will be used: {label}"
                         )
        parameter_dict["label"] = label
    return parameter_dict


def _check_or_set_dataset_type(
        app,
        parameter_dict=None,
        dataset_type=None,
        dataset_file=None
):
    """Check for missing dataset type.

    Args:
        app (ApplicationContext): the click cli app context.
        parameter_dict (dict): The dictionary of given dataset parameters
        dataset_type (str): The dataset type ID.
        dataset_file (str): The path of the local file.

    Returns:
        A dataset type based the given file names.
    """
    if dataset_file:
        source = os.path.basename(dataset_file)
    else:
        source = ""
    if "file" in parameter_dict:
        target = parameter_dict["file"]
    else:
        target = ""
    suggestions = (
        (".ttl", "file"),
        (".csv", "csv"),
        (".xlsx", "excel"),
        (".xml", "xml"),
        (".json", "json"),
        (".orc", "orc"),
        (".zip", "multiCsv")
    )
    if not dataset_type:
        for check, type_ in suggestions:
            if source.endswith(check) or target.endswith(check):
                dataset_type = type_
        if not dataset_type:
            raise ValueError("Missing parameter. Please specify a dataset "
                             "type with '--type'.")
        app.echo_warning(
            "Missing dataset type (--type) - based on the used file name, "
            f"this type is assumed: {dataset_type}"
        )
    return dataset_type


def _show_parameter_list(app, dataset_type=None):
    """Output the parameter list for a given dataset type.

    Args:
        app (ApplicationContext): the click cli app context.
        dataset_type (dict): The type from which the parameters are listed.
    """
    plugin = get_task_plugin_description(dataset_type)
    properties = plugin["properties"]
    required_properties = plugin["required"]
    table = []
    for key in properties:
        if key in required_properties:
            parameter = key + " *"
            description = "(Required) " + properties[key]["description"]
        else:
            parameter = key
            description = properties[key]["description"]
        row = [
            parameter,
            description,
        ]
        table.append(row)
    # metadata always on top, then sorted by key
    table = sorted(table, key=lambda k: k[0].lower())
    table = completion.add_metadata_parameter(table)
    app.echo_info_table(
        table,
        headers=["Parameter", "Description"]
    )


def _show_type_list(app):
    """Output the list of dataset types.

    Args:
        app (ApplicationContext): the click cli app context.
    """
    plugins = get_task_plugins()
    table = []
    for plugin_id in plugins:
        plugin = plugins[plugin_id]
        if plugin["taskType"] == "Dataset":
            id_ = plugin_id
            title = plugin["title"]
            description = plugin["description"].partition("\n")[0]
            row = [
                id_,
                f"{title}: {description}",
            ]
            table.append(row)
    app.echo_info_table(
        table,
        headers=["Dataset Type", "Description"],
        sort_column=1
    )


def _check_or_select_project(app, project_id=None):
    """Check for given project, select the first one if there is only one.

    Args:
        app (ApplicationContext): the click cli app context.
        project_id (str): The project ID.

    Raises:
        ValueError: if no projects available.
        ValueError: if more than one project is.

    Returns:
        Maybe project_id if there was no project_id before.
    """
    if project_id is not None:
        return project_id

    projects = get_projects()
    if len(projects) == 1:
        project_name = projects[0]["name"]
        app.echo_warning(
            "Missing project (--project) - since there is only one project, "
            f"this is selected: {project_name}"
        )
        return project_name

    if len(projects) == 0:
        raise ValueError(
            "There are no projects available. "
            "Please create a project with 'cmemc project create'."
        )

    # more than one project
    raise ValueError(
        "There is more than one project available so you need to "
        "specify the project with '--project'."
    )


def _check_or_select_dataset_type(app, dataset_type):
    """Test type and return plugin.

    Args:
        app (ApplicationContext): the click cli app context.
        dataset_type (str): A dataset type

    Raises:
        ValueError: If type is not known

    Returns:
        A tuple of dataset_type and corresponding plugin description (dict)
    """
    try:
        app.echo_debug(f"check type {dataset_type}")
        plugin = get_task_plugin_description(dataset_type)
        return dataset_type, plugin
    except requests.exceptions.HTTPError as error:
        raise ValueError(
            f"Unknown dataset type: {dataset_type}."
        ) from error


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--project", "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids,
    help="The project, from which you want to list the datasets. "
         "Project IDs can be listed with the 'project list' command."
)
@click.option(
    "--filter", "filter_",
    type=(str, str),
    multiple=True,
    autocompletion=completion.dataset_list_filter,
    help=DATASET_FILTER_HELP_TEXT,
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON objects of dataset search API response."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only dataset identifier and no labels or other meta data. "
         "This is useful for piping the ids into other cmemc commands."
)
@click.pass_obj
def list_command(app: ApplicationContext, project_id, filter_, raw, id_only):
    """List available datasets.

    Outputs a list of datasets IDs which can be used as reference for
    the dataset create and delete commands.
    """
    if project_id:
        app.echo_warning(
            "Option '--project' is deprecated and will be removed. "
            "Please use '--filter project XXX' instead."
        )
    results = list_items(item_type="dataset", project=project_id)
    datasets = results["results"]
    for _ in filter_:
        filter_type, filter_name = _
        datasets = _get_datasets_filtered(datasets, filter_type, filter_name)

    if raw:
        app.echo_info_json(datasets)
    elif id_only:
        for _ in datasets:
            app.echo_info(_["projectId"] + ":" + _["id"])
    else:
        table = []
        for _ in datasets:
            row = [
                _["projectId"] + ":" + _["id"],
                _["pluginId"],
                _["label"],
            ]
            table.append(row)
        app.echo_info_table(
            table,
            headers=["Dataset ID", "Type", "Label"],
            sort_column=2
        )


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Delete all datasets. "
         "This is a dangerous option, so use it with care.",
)
@click.option(
    "--project", "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids,
    help="In combination with the '--all' flag, this option allows for "
         "deletion of all datasets of a certain project. The behaviour is "
         "similar to the 'dataset list --project' command."
)
@click.argument(
    "dataset_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.dataset_ids
)
@click.pass_obj
def delete_command(app, project_id, all_, dataset_ids):
    """Delete datasets.

    This deletes existing datasets in integration projects from Corporate
    Memory. Datasets will be deleted without prompting!
    Dataset resources will not be deleted.

    Example: cmemc dataset delete my_project:my_dataset

    Datasets can be listed by using the 'cmemc dataset list' command.
    """
    if dataset_ids == () and not all_:
        raise ValueError("Either specify at least one dataset ID "
                         + "or use the --all option to delete all datasets.")

    if all_:
        # in case --all is given, a list of datasets is fetched
        dataset_ids = []
        datasets = list_items(item_type="dataset", project=project_id)
        for _ in datasets["results"]:
            dataset_ids.append(_["projectId"] + ":" + _["id"])

    count = len(dataset_ids)
    current = 1
    for _ in dataset_ids:
        app.echo_info(
            f"Delete dataset {current}/{count}: {_} ... ",
            nl=False
        )
        project_part, dataset_part = _validate_and_split_dataset_id(_)
        app.echo_debug(
            f"Project ID is {project_part}, dataset ID is {dataset_part}"
        )
        delete_dataset(project_part, dataset_part)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="download")
@click.argument(
    "dataset_id",
    type=click.STRING,
    autocompletion=completion.dataset_ids
)
@click.argument(
    "output_path",
    required=True,
    type=click.Path(
        allow_dash=True,
        dir_okay=False,
        writable=True
    )
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace existing files. This is a dangerous option, "
         "so use it with care.",
)
@click.pass_obj
def download_command(app, dataset_id, output_path, replace):
    """Download the resource file of a dataset.

    This command downloads the file resource of a dataset to your local
    file system or to standard out (-).
    Note that this is not possible for dataset types such as
    Knowledge Graph (eccencaDataplatform) or SQL endpoint (sqlEndpoint).

    Without providing an output path, the output file name will be the
    same as the remote file resource.

    Datasets can be listed by using the 'cmemc dataset list' command.
    """
    app.echo_debug(
        f"Dataset ID is {dataset_id}; "
        f"output path is {click.format_filename(output_path)}. "
        f"replace is {replace}."
    )
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    try:
        file = project['data']['parameters']['file']
    except KeyError as no_file_resource:
        raise ValueError(
            f"The dataset {dataset_id} has no associated file resource."
        ) from no_file_resource
    if os.path.exists(output_path) and replace is not True:
        raise ValueError(
            f"Target file {click.format_filename(output_path)} already "
            "exists. Use --replace in case you want to replace it."
        )
    with get_resource_response(project_part, file) as response:
        # if piping file to stdout, no info messages
        if output_path != "-":
            app.echo_info(
                f"Download resource {file} of dataset {dataset_id} to file "
                f"{click.format_filename(output_path)} ... ",
                nl=False
            )
        with click.open_file(output_path, "wb") as resource_file:
            for chunk in response.iter_content(chunk_size=8192):
                resource_file.write(chunk)
            # if piping file to stdout, no info messages
        if output_path != "-":
            app.echo_success("done")


@click.command(cls=CmemcCommand, name="upload")
@click.argument(
    "dataset_id",
    type=click.STRING,
    autocompletion=completion.dataset_ids
)
@click.argument(
    "input_path",
    required=True,
    autocompletion=completion.dataset_files,
    type=click.Path(
        allow_dash=True,
        dir_okay=False,
        writable=True
    )
)
@click.pass_obj
def upload_command(app, dataset_id, input_path):
    """Upload a resource file to a dataset.

    This command uploads a file to a dataset.
    The content of the uploaded file replaces the remote file resource.
    The name of the remote file resource is not changed.

    Warning: If the remote file resource is used in more than one dataset,
    the other datasets are also affected by this command.

    Warning: The content of the uploaded file is not tested, so uploading
    a json file to an xml dataset will result in errors.

    Datasets can be listed by using the 'cmemc dataset list' command.

    Example: cmemc dataset upload cmem:my-dataset new-file.csv
    """
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    try:
        remote_file_name = project["data"]["parameters"]["file"]
    except KeyError as error:
        raise ValueError(
            f"Dataset {dataset_id} has no attached file resource to replace."
        ) from error

    _upload_file_resource(
        app=app,
        project_id=project_part,
        local_file_name=input_path,
        remote_file_name=remote_file_name,
        replace=True
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument(
    "dataset_id",
    type=click.STRING,
    autocompletion=completion.dataset_ids
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def inspect_command(app, dataset_id, raw):
    """Display meta data of a dataset."""
    app.echo_debug(f"Dataset ID is {dataset_id}")
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    if raw:
        app.echo_info_json(project)
    else:
        table = struct_to_table(project)
        app.echo_info_table(
            table,
            headers=["Key", "Value"],
            sort_column=0
        )


# pylint: disable-msg=too-many-locals,too-many-arguments
@click.command(cls=CmemcCommand, name="create")
@click.argument(
    "DATASET_FILE",
    required=False,
    autocompletion=completion.dataset_files,
    type=click.Path(
        allow_dash=False,
        readable=True,
        exists=True
    )
    # TODO: Allow dash here / enable creation of file resources from stdin.
)
@click.option(
    "--type", "-t", "dataset_type",
    multiple=False,
    type=click.STRING,
    autocompletion=completion.dataset_types,
    help="The dataset type of the dataset to create. Example types are 'csv',"
         "'json' and 'eccencaDataPlatform' (-> Knowledge Graph)."
)
@click.option(
    "--project", "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids,
    help="The project, where you want to create the dataset in. If there is "
         "only one project in the workspace, this option can be omitted."
)
@click.option(
    "--parameter", "-p",
    type=(str, str),
    autocompletion=completion.dataset_parameter,
    multiple=True,
    help="A set of key/value pairs. Each dataset type has different "
         "parameters (such as charset, arraySeparator, ignoreBadLines, ...). "
         "In order to get a list of possible parameter, use the"
         "'--help-parameter' option."
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace remote file resources in case there "
         "already exists a file with the same name."
)
@click.option(
    "--id", "dataset_id",
    type=click.STRING,
    help="The dataset ID of the dataset to create. The dataset ID will be "
         "automatically created in case it is not present."
)
@click.option(
    "--help-types",
    is_flag=True,
    help="Lists all possible dataset types on given Corporate Memory "
         "instance. Note that this option already needs access to the "
         "instance."
)
@click.option(
    "--help-parameter",
    is_flag=True,
    help="Lists all possible (optional and mandatory) parameter for a dataset "
         "type. Note that this option already needs access to the instance."
)
@click.pass_obj
def create_command(
        app, dataset_file, replace,
        project_id, dataset_id, dataset_type, parameter,
        help_parameter, help_types
):
    """Create a dataset.

    Datasets are created in projects and can have associated file
    resources. Each dataset has a type (such as 'csv') and a list of
    parameter which can change or specify the dataset behaviour.

    To get more information on possible dataset types and parameter on these
    types, use the '--help-types' and '--help-parameter' options.

    Example: cmemc dataset create --project my-project --type csv my-file.csv
    """
    if help_types:
        _show_type_list(app)
        return

    # transform the parameter list of tuple to a dictionary
    parameter_dict = {}
    for key, value in parameter:
        parameter_dict[key] = value

    dataset_type = _check_or_set_dataset_type(
        app,
        parameter_dict=parameter_dict,
        dataset_type=dataset_type,
        dataset_file=dataset_file
    )

    if help_parameter:
        _show_parameter_list(app, dataset_type=dataset_type)
        return

    dataset_type, plugin = _check_or_select_dataset_type(app, dataset_type)

    parameter_dict = _extend_parameter_with_metadata(
        app,
        parameter_dict=parameter_dict,
        dataset_type=dataset_type,
        dataset_file=dataset_file
    )

    project_id = _check_or_select_project(app, project_id)

    # file required but not given
    if "file" in plugin["required"] \
            and not dataset_file \
            and "file" not in parameter_dict:
        raise ValueError(
            f"The dataset type {dataset_type} is file based, so you need "
            "to specify a file with the create command."
        )

    # file required and given
    # dataset_file = file path from the command line
    # parameter_dict["file"] = local name in DI
    if "file" in plugin["required"] and dataset_file:
        # add file parameter for the project if needed
        if "file" not in parameter_dict:
            parameter_dict["file"] = os.path.basename(dataset_file)
        _upload_file_resource(
            app=app,
            project_id=project_id,
            local_file_name=dataset_file,
            remote_file_name=parameter_dict["file"],
            replace=replace
        )

    # create dataset resource
    app.echo_info(
        f"Create a new dataset {project_id}:",
        nl=False
    )
    created_dataset = create_dataset(
        dataset_id=dataset_id, project_id=project_id,
        dataset_type=dataset_type, parameter=parameter_dict,
        metadata=_get_metadata_out_of_parameter(parameter_dict)
    )
    returned_id = json.loads(created_dataset)["id"]
    app.echo_info(f"{returned_id} ... ", nl=False)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "dataset_ids",
    nargs=-1,
    required=True,
    type=click.STRING,
    autocompletion=completion.dataset_ids
)
@click.pass_obj
def open_command(app, dataset_ids):
    """Open datasets in the browser.

    With this command, you can open a dataset in the workspace in
    your browser.

    The command accepts multiple dataset IDs which results in
    opening multiple browser tabs.
    """
    dataset_urls = {}
    for _ in list_items(item_type="dataset")["results"]:
        dataset_id = _["projectId"] + ":" + _["id"]
        url = _["itemLinks"][0]["path"]
        dataset_urls[dataset_id] = url
    for _ in dataset_ids:
        if _ in dataset_urls:
            full_url = get_cmem_base_uri() + dataset_urls[_]
            app.echo_debug(f"Open {_}: {full_url}")
            click.launch(full_url)
        else:
            raise ValueError(f"Dataset '{_}' not found.")


@click.group(cls=CmemcGroup)
def dataset():
    """List, create, delete, inspect, up-/download or open datasets.

    This command group allows for managing workspace datasets as well as
    dataset file resources. Datasets can be created and deleted.
    File resources can be uploaded and downloaded.
    Details of dataset parameter can be listed with inspect.

    Datasets are identified with a combined key of the project ID and the
    project internal dataset ID (e.g: my-project:my-dataset). To get a list
    of datasets, use the list command.
    """


dataset.add_command(list_command)
dataset.add_command(delete_command)
dataset.add_command(download_command)
dataset.add_command(upload_command)
dataset.add_command(inspect_command)
dataset.add_command(create_command)
dataset.add_command(open_command)
dataset.add_command(resource)



"""admin commands for cmem command line interface."""
import click

import jwt

from cmem.cmemc.cli.utils import (
    get_graphs_as_dict, struct_to_table
)
from cmem.cmempy.api import (
    get_access_token, get_token
)
from cmem.cmempy.config import (
    get_cmem_base_uri
)
from cmem.cmempy.health import (
    di_is_healthy, dp_is_healthy,
    get_di_version, get_dm_version, get_dp_version,
    get_shape_catalog_version
)
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.commands.store import (
    bootstrap_command,
    showcase_command,
    store
)
from cmem.cmemc.cli.commands.metrics import metrics
from cmem.cmemc.cli.commands.workspace import workspace


@click.command(cls=CmemcCommand, name="status")
@click.pass_obj
def status_command(app):
    """Output health and version information.

    This command outputs version and health information of the
    selected deployment. If the version information can not be retrieved,
    UNKNOWN is shown if the endpoint is not available or ERROR is shown,
    if the endpoints returns an error.

    In addition to that, this command warns you if the
    target version of your cmemc client is newer than the version of your
    backend and if the ShapeCatalog has a different version then your
    DataPlatform component.

    To get status information of all configured
    deployments use this command in combination with parallel:

    cmemc config list | parallel --ctag cmemc -c {} admin status
    """
    app.check_versions()
    app.echo_result(get_cmem_base_uri())
    # check DI
    app.echo_result("- DataIntegration ", nl=False)
    app.echo_result(get_di_version() + " ... ", nl=False)
    if di_is_healthy():
        app.echo_success("UP")
    else:
        app.echo_error("DOWN")
    # check DP
    app.echo_result("- DataPlatform ", nl=False)
    dp_version = get_dp_version()
    app.echo_result(dp_version + " ... ", nl=False)
    if dp_is_healthy():
        app.echo_success("UP")
    else:
        app.echo_error("DOWN")
    # check shape catalog
    shapes_version = get_shape_catalog_version()
    app.echo_result("-- ShapesCatalog ", nl=False)
    app.echo_result(shapes_version + " ... ", nl=False)
    if "https://vocab.eccenca.com/shacl/" not in get_graphs_as_dict():
        app.echo_error("DOWN")
    else:
        app.echo_success("UP")
    if shapes_version not in (dp_version, "UNKNOWN"):
        app.echo_warning(
            "Your ShapeCatalog version does not match your DataPlatform "
            "version. Please consider updating your bootstrap data."
        )
    # check DM
    app.echo_result("- DataManager ", nl=False)
    dm_version = get_dm_version()
    app.echo_result(dm_version + " ... ", nl=False)
    if dm_version == "ERROR":
        app.echo_error("DOWN")
    else:
        app.echo_success("UP")


@click.command(cls=CmemcCommand, name="token")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON. Note that this option will always try to fetch "
         "a new JSON token response. In case you are working with "
         "OAUTH_GRANT_TYPE=prefetched_token, this may lead to an error."
)
@click.option(
    "--decode",
    is_flag=True,
    help="Decode the access token and outputs the raw JSON. Note that the "
         "access token is only decoded and esp. not validated."
)
@click.pass_obj
def token_command(app, raw, decode):
    """Fetch and output an access token.

    This command can be used to check for correct authentication as well as
    to use the token with wget / curl or similar standard tools:

    Example Usage: curl -H "Authorization: Bearer $(cmemc -c my admin token)"
    $(cmemc -c my config get DP_API_ENDPOINT)/api/custom/slug

    Please be aware that this command can reveal secrets, which you do not
    want to have in log files or on the screen.
    """
    # Note:
    # - get_access_token returns the token string which is maybe from conf
    # - get_token fetches a new token incl. envelope from keycloak

    if decode:
        token = get_access_token()
        decoded_token = jwt.decode(token, options={"verify_signature": False})
        if raw:
            app.echo_info_json(decoded_token)
        else:
            table = struct_to_table(decoded_token)
            app.echo_info_table(
                table,
                headers=["Key", "Value"],
                sort_column=0
            )
    else:
        if raw:
            app.echo_info_json(get_token())
        else:
            token = get_access_token()
            app.echo_info(token)


@click.group(cls=CmemcGroup)
def admin():
    """Import bootstrap data, backup/restore workspace or get status.

    This command group consists of commands for setting up and
    configuring eccenca Corporate Memory.
    """


admin.add_command(showcase_command)
admin.add_command(bootstrap_command)
admin.add_command(status_command)
admin.add_command(token_command)
admin.add_command(metrics)
admin.add_command(workspace)
admin.add_command(store)



"""DataIntegration python management commands."""
import sys
from re import match

import click

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.context import ApplicationContext
from cmem.cmempy.workspace.python import (
    install_package_by_file, install_package_by_name,
    list_packages, list_plugins,
    uninstall_package, update_plugins
)


def _get_package_id(module_name: str) -> str:
    """Return package identifier."""
    return module_name.split('.')[0].replace('_', '-')


def _looks_like_a_package(package: str) -> bool:
    """Check if a string looks like a package requirement string."""
    if match(r"^[a-zA-Z0-9]+(-[a-zA-Z0-9]+)*((==|<=|>=|>|<).*)?$", package):
        return True
    return False


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "PACKAGE",
    autocompletion=completion.installable_packages,
    type=click.Path(
        readable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def install_command(app: ApplicationContext, package):
    """Install a python package to the workspace.

    This command is basically a 'pip install' in the remote python
    environment.

    You can install a package by uploading a source distribution
    .tar.gz file, or by uploading a build distribution .whl file, or by
    specifying a package name, more precisely, a pip requirement
    specifier with a package name available on pypi.org
    (e.g. 'requests==2.27.1').
    """
    app.echo_info(
        f"Install package {package} ... ",
        nl=False
    )
    try:
        response = install_package_by_file(package_file=package)
    except FileNotFoundError as error:
        if not _looks_like_a_package(package):
            raise ValueError(
                f"{package} does not look like a package name or requirement "
                "string, and a file with this name also does not exists."
            ) from error
        response = install_package_by_name(package_name=package)
    for output_line in response["standardOutput"].splitlines():
        app.echo_debug(output_line)

    if response["success"]:
        app.echo_success("done")
        if len(response["errorOutput"].strip()) > 0:
            app.echo_warning(response["errorOutput"])
    else:
        app.echo_error("error")
        app.echo_debug(response["standardOutput"])
        if len(response["errorOutput"].strip()) > 0:
            app.echo_error(response["errorOutput"])
    app.echo_debug("Updated Plugins: " + str(update_plugins()))


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument(
    "PACKAGE_NAME",
    autocompletion=completion.installed_package_names
)
@click.pass_obj
def uninstall_command(app: ApplicationContext, package_name):
    """Uninstall a python package from the workspace.

    This command is basically a 'pip uninstall' in the remote
    python environment.
    """
    app.echo_info(
        f"Uninstall package {package_name} ... ",
        nl=False
    )
    packages = list_packages()
    app.echo_debug(packages)
    if package_name not in [package["name"] for package in packages]:
        app.echo_error("not installed")
        sys.exit(1)
    response = uninstall_package(package_name=package_name)
    for output_line in response["standardOutput"].splitlines():
        app.echo_debug(output_line)
    if response["success"]:
        app.echo_success("done")
        if len(response["errorOutput"].strip()) > 0:
            app.echo_warning(response["errorOutput"])
    else:
        app.echo_error("error")
        app.echo_debug(response.content.decode())
    app.echo_debug("Updated Plugins: " + str(update_plugins()))


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only package identifier. "
         "This is useful for piping the IDs into other commands."
)
@click.pass_obj
def list_command(app: ApplicationContext, raw, id_only):
    """List installed python packages.

    This command is basically a 'pip list' in the remote python environment.

    It outputs a table of python package identifiers with version information.
    """
    packages = list_packages()
    if raw:
        app.echo_info_json(packages)
        return
    if id_only:
        for package in packages:
            app.echo_info(package["name"])
        return
    table = []
    for package in packages:
        table.append((
            package["name"],
            package["version"]
        ))
    app.echo_info_table(
        table,
        headers=["Name", "Version"],
        sort_column=0
    )


@click.command(cls=CmemcCommand, name="list-plugins")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only plugin identifier."
)
@click.option(
    "--package-id-only",
    is_flag=True,
    help="Lists only plugin package identifier."
)
@click.pass_obj
def list_plugins_command(
        app: ApplicationContext, raw, id_only, package_id_only
):
    """List installed workspace plugins.

    This commands lists all discovered plugins.
    Note that the plugin discovery is limited to specific packages.
    """
    raw_output = list_plugins()
    try:
        # 22.1.1 DI output
        plugins = raw_output["plugins"]
    except TypeError:
        # 22.1 DI output
        plugins = raw_output

    if raw:
        app.echo_info_json(plugins)
        return
    if id_only:
        for plugin in sorted(plugins, key=lambda k: k["id"].lower()):
            app.echo_info(plugin["id"])
        return
    if package_id_only:
        for package_id in sorted(
            {_get_package_id(plugin["moduleName"]) for plugin in plugins}
        ):
            app.echo_info(package_id)
        return
    table = []
    for plugin in sorted(plugins, key=lambda k: k["id"].lower()):
        table.append((
            plugin["id"],
            _get_package_id(plugin["moduleName"]),
            plugin["pluginType"],
            plugin["label"],
        ))
    app.echo_info_table(
        table,
        headers=["ID", "Package ID", "Type", "Label"],
        sort_column=0
    )
    if "error" in raw_output:
        app.echo_error(raw_output["error"])


@click.group(cls=CmemcGroup)
def python():
    """List, install, or uninstall python packages.

    Python packages are used to extend the DataIntegration workspace
    with python plugins. To get a list of installed packages, execute the
    list command.

    Warning: Installing packages from unknown sources is not recommended.
    Plugins are not verified for malicious code.
    """


python.add_command(install_command)
python.add_command(uninstall_command)
python.add_command(list_command)
python.add_command(list_plugins_command)



"""query commands for cmem command line interface."""
import json
import re
from hashlib import sha1
from shutil import get_terminal_size
from json import JSONDecodeError, load
from time import sleep, time
from typing import Dict, List
from uuid import uuid4

import click

from requests import HTTPError

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.utils import extract_error_message, struct_to_table
from cmem.cmempy.queries import (
    QUERY_CATALOG,
    SparqlQuery,
    get_query_status
)

QUERY_FILTER_TYPES = ['status', 'slower-than', 'type', 'regex']
QUERY_FILTER_HELP_TEXT = (
    "Filter queries based on execution status and time. "
    f"First parameter --filter CHOICE can be one of {str(QUERY_FILTER_TYPES)}"
    ". The second parameter is based on CHOICE, e.g. int in case of"
    " slower-than, or a regular expression string."
)


class ReplayStatistics:
    """Capture and calculate statistics of a query replay command run."""

    # pylint: disable=too-many-instance-attributes
    runId: str = str(uuid4())
    query_minimum: int = None
    query_maximum: int = None
    loop_count: int = 0
    loop_durations: Dict[str, int] = {}
    query_durations: Dict[str, List[int]] = {}
    current_loop_key: str = None
    total_duration: int = 0
    query_count: int = 0
    error_count: int = 0
    query_average: float = None
    catalog = QUERY_CATALOG
    app = None

    def __init__(self, app, label: str = None):
        """Initialize instance."""
        self.app = app
        self.label = label

    def init_loop(self) -> str:
        """Initialize a new loop and reset the loop counts/values."""
        loop_key = str(uuid4())
        self.current_loop_key = loop_key
        self.loop_durations[loop_key] = 0
        self.query_durations[loop_key] = []
        self.loop_count += 1
        self.app.echo_debug(f"Loop {self.loop_count} started: {loop_key}")
        return loop_key

    def _init_query(self, query_: dict) -> SparqlQuery:
        """Initialize query from dict."""
        try:
            if "iri" in query_.keys():
                iri = query_["iri"]
                catalog_entry = self.catalog.get_query(iri)
                if catalog_entry is None:
                    raise ValueError(
                        f"measure_query - query {iri} is not in catalog."
                    )
                return catalog_entry
            query_string = query_["queryString"]
            return SparqlQuery(text=query_string)
        except KeyError as error:
            raise ValueError(
                "measure_query - given input dict has no queryString key."
            ) from error

    def _update_statistic_on_success(self, duration: int):
        """Update statistics and counters."""
        self.query_durations[self.current_loop_key].append(duration)
        if self.query_minimum is None or duration < self.query_minimum:
            self.query_minimum = duration
        if self.query_maximum is None or duration > self.query_maximum:
            self.query_maximum = duration
        self.total_duration += duration
        self.loop_durations[self.current_loop_key] += duration
        self.query_average = self.total_duration / self.query_count

    def measure_query_duration(self, query_: dict) -> dict:
        """Execute a query and measure the duration."""
        # create the query object
        executed_query = self._init_query(query_)

        # update and return the query object
        if "id" not in query_.keys():
            # create a UUID4, if needed
            query_["id"] = str(uuid4())
        # always re-hash
        query_["queryStringSha1Hash"] = sha1(  # nosec
            executed_query.text.encode('utf-8')
        ).hexdigest()
        if "queryString" not in query_.keys():
            # add queryString for catalog queries
            query_["queryString"] = executed_query.text
        if "iri" in query_.keys():
            # use the full IRI in case short one is given
            query_["iri"] = executed_query.url
        if "iri" in query_.keys() and "label" not in query_.keys():
            # extend with label if possible (and needed)
            query_["label"] = executed_query.label
        if "replays" not in query_.keys():
            # create replays object if needed
            query_["replays"] = []

        # init replay object
        this_replay = {
            "runId": str(self.runId),
            "loopId": str(self.current_loop_key),
            "replayId": str(uuid4()),
        }
        if self.label is not None:
            this_replay["runLabel"] = self.label

        # execute and measure the query
        try:
            start = round(time() * 1000)
            executed_query.get_results(replace_placeholder=False)
            self.query_count += 1
            end = round(time() * 1000)
            duration = end - start

            this_replay["executionStarted"] = str(start)
            this_replay["executionFinished"] = str(end)
            this_replay["executionTime"] = str(duration)

            # update statistics and counters
            self._update_statistic_on_success(duration)
        except HTTPError as error:
            self.error_count += 1
            this_replay["executionError"] = extract_error_message(error)

        query_["replays"].append(this_replay)
        self.app.echo_debug(
            f"Query {self.query_count + self.error_count} "
            f"executed: {this_replay['replayId']}"
        )
        return query_

    def create_output(self):
        """Create the structure for the output commands."""
        output = {}
        for key, value in vars(self).items():
            if key not in ("current_loop_key", "app"):
                # ignore some object vars on output
                output[key] = value
        loop_average = 0
        loop_minimum = None
        loop_maximum = None
        for loop_duration in self.loop_durations.values():
            loop_average += loop_duration
            if loop_minimum is None or loop_duration < loop_minimum:
                loop_minimum = loop_duration
            if loop_maximum is None or loop_duration > loop_maximum:
                loop_maximum = loop_duration
        output["loop_minimum"] = loop_minimum
        output["loop_maximum"] = loop_maximum
        output["loop_average"] = loop_average / len(self.loop_durations)
        return output

    def output_table(self):
        """Output a table of the statistic values."""
        table = struct_to_table(self.create_output())
        self.app.echo_info_table(
            table,
            headers=["Key", "Value"],
            sort_column=0
        )

    def output_json(self):
        """Output json of the statistic values."""
        self.app.echo_info_json(self.create_output())


def _get_queries_filtered_by_type(queries, query_type: str):
    """Get queries filter by query type.

    Args:
        query_type (str): the type of the query

    Returns:
        filtered list of status reports

    Raises:
        ValueError
    """
    filtered_queries = []
    try:
        for _ in queries:
            if _["type"].lower() == query_type.lower():
                filtered_queries.append(_)
    except KeyError as error:
        raise NotImplementedError(
            "Filtering queries by type needs DataPlatform >= 22.1"
        ) from error
    return filtered_queries


def _get_queries_filtered_by_slower_than(queries, time_ms):
    """Get queries filter by slower-than x ms filter.

    Args:
        time_ms (int):time in milliseconds

    Returns:
        filtered list of status reports

    Raises:
        ValueError
    """
    try:
        time_ms = int(time_ms)
    except ValueError as error:
        raise ValueError("This filter needs an integer as parameter.") \
            from error
    filtered_queries = []
    for _ in queries:
        if _["executionTime"] is not None and _["executionTime"] > time_ms:
            filtered_queries.append(_)
    return filtered_queries


def _get_queries_filtered_by_status(queries, status):  # noqa: C901
    """Get queries filter by a status.

    Args:
        status (str): one of error, running, finished

    Returns:
        filtered list of status reports

    Raises:
        ValueError
    """
    # check for correct filter names (filter ids is used internally only)

    valid_filter_values = ("running", "finished", "error")
    if status not in valid_filter_values:
        raise ValueError(
            f"This filter needs one of {valid_filter_values} as parameter."
        )
    filtered_queries = []
    if status == "running":
        for _ in queries:
            if _["executionTime"] is None:
                filtered_queries.append(_)
    elif status == "finished":
        for _ in queries:
            if _["executionTime"] is not None:
                filtered_queries.append(_)
    elif status == "error":
        for _ in queries:
            if _["error"] is not None:
                filtered_queries.append(_)
    return filtered_queries


def _get_queries_filtered_by_uuid(queries, uuid):
    """Get query status by uuid.

    Args:
        uuid (str): query status uuid

    Returns:
        status dict
    """
    filtered_queries = []
    for _ in queries:
        if _["id"] == uuid:
            filtered_queries.append(_)
    return filtered_queries


def _get_queries_filtered(queries, filter_name, filter_value):
    """Get queries filtered according to filter name and value.

    Args:
        filter_name (str): one of "status" or "slower-than"
        filter_value (str|int): value according to filter

    Returns:
        list of filtered queries from the get_query_status API call

    Raises:
        ValueError
    """
    # filter by uuid
    if filter_name == "uuid":
        return _get_queries_filtered_by_uuid(queries, filter_value)
    # check for correct filter names (filter ids is used internally only)
    if filter_name not in QUERY_FILTER_TYPES:
        raise ValueError(
            f"{filter_name} is an unknown filter name. "
            f"Use one of {QUERY_FILTER_TYPES}."
        )
    # filter by status
    if filter_name == "status":
        return _get_queries_filtered_by_status(queries, filter_value)
    # filter by slower-than
    if filter_name == "slower-than":
        return _get_queries_filtered_by_slower_than(queries, filter_value)
    # filter by type
    if filter_name == "type":
        return _get_queries_filtered_by_type(queries, filter_value)
    # filter by regex
    if filter_name == "regex":
        return [_ for _ in queries
                if re.search(filter_value, _["queryString"])]
    # default is unfiltered
    return queries


def _output_query_status_details(app, status_dict):
    """Output key/value table as well as query string of a query.

    Args:
        status_dict (dict): The dict from the query status list.
    """
    table = []
    for key in status_dict.keys():
        if key != "queryString":
            row = [
                key,
                str(status_dict[key])
            ]
            table.append(row)
    app.echo_info_table(
        table,
        headers=["Key", "Value"],
        sort_column=0
    )
    app.echo_info("")
    app.echo_info_sparql(status_dict["queryString"])


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only query identifier and no labels or other meta data. "
         "This is useful for piping the ids into other cmemc commands."
)
@click.pass_obj
def list_command(app, id_only):
    """List available queries from the catalog.

    Outputs a list of query URIs which can be used as reference for
    the query execute command.
    """
    queries = QUERY_CATALOG.get_queries().items()
    if id_only:
        # sort dict by short_url - https://docs.python.org/3/howto/sorting.html
        for _, sparql_query in sorted(
                queries, key=lambda k: k[1].short_url.lower()
        ):
            app.echo_info(
                sparql_query.short_url
            )
    else:
        table = []
        for _, sparql_query in queries:
            row = [
                sparql_query.short_url,
                sparql_query.query_type,
                ','.join(sparql_query.get_placeholder_keys()),
                sparql_query.label
            ]
            table.append(row)
        app.echo_info_table(
            table,
            headers=["Query URI", "Type", "Placeholder", "Label"],
            sort_column=3
        )


# pylint: disable-msg=too-many-locals,too-many-arguments
@click.command(cls=CmemcCommand, name="execute")
@click.argument(
    "QUERIES",
    nargs=-1,
    required=True,
    autocompletion=completion.remote_queries_and_sparql_files
)
@click.option(
    "--accept",
    default="default",
    show_default=True,
    help="Accept header for the HTTP request(s). Setting this to 'default' "
         "means that cmemc uses an appropriate accept header for terminal "
         "output (text/csv for tables, text/turtle for graphs, * otherwise). "
         "Please refer to the Corporate Memory system manual for a list "
         "of accepted mime types."
)
@click.option(
    "--no-imports",
    is_flag=True,
    help="Graphs which include other graphs (using owl:imports) will be "
         "queried as merged overall-graph. This flag disables this "
         "default behaviour. The flag has no effect on update queries."
)
@click.option(
    "--base64",
    is_flag=True,
    help="Enables base64 encoding of the query parameter for the "
         "SPARQL requests (the response is not touched). "
         "This can be useful in case there is an aggressive firewall between "
         "cmemc and Corporate Memory."
)
@click.option(
    "--parameter", "-p",
    type=(str, str),
    autocompletion=completion.placeholder,
    multiple=True,
    help="In case of a parameterized query (placeholders with the '{{key}}' "
         "syntax), this option fills all placeholder with a given value "
         "before the query is executed."
         "Pairs of placeholder/value need to be given as a tuple 'KEY VALUE'. "
         "A key can be used only once."
)
@click.option(
    "--limit",
    type=int,
    help="Override or set the LIMIT in the executed SELECT query. Note that "
         "this option will never give you more results than the LIMIT given "
         "in the query itself."
)
@click.option(
    "--offset",
    type=int,
    help="Override or set the OFFSET in the executed SELECT query."
)
@click.option(
    "--distinct",
    is_flag=True,
    help="Override the SELECT query by make the result set DISTINCT."
)
@click.option(
    "--timeout",
    type=int,
    help="Set max execution time for query evaluation (in milliseconds)."
)
@click.pass_obj
def execute_command(
        app, queries, accept, no_imports, base64, parameter,
        limit, offset, distinct, timeout
):
    """Execute queries which are loaded from files or the query catalog.

    Queries are identified either by a file path, a URI from the query
    catalog, or a shortened URI (qname, using a default namespace).

    If multiple queries are executed one after the other, the first failing
    query stops the whole execution chain.

    Limitations: All optional parameters (e.g. accept, base64, ...) are
    provided for ALL queries in an execution chain. If you need different
    parameters for each query in a chain, run cmemc multiple times and use
    the logical operators && and || of your shell instead.
    """
    # pylint: disable=too-many-arguments
    placeholder = {}
    for key, value in parameter:
        if key in placeholder:
            raise ValueError(
                "Parameter can be given only once, "
                f"Value for '{key}' was given twice."
            )
        placeholder[key] = value
    app.echo_debug("Parameter: " + str(placeholder))
    for file_or_uri in queries:
        app.echo_debug(
            f"Start of execution: {file_or_uri} with "
            f"placeholder {placeholder}"
        )
        executed_query = QUERY_CATALOG.get_query(
            file_or_uri, placeholder=placeholder
        )
        if executed_query is None:
            raise ValueError(
                f"{file_or_uri} is neither a (readable) file nor a query URI."
            )
        app.echo_debug(
            f"Execute ({executed_query.query_type}): "
            f"{executed_query.label} < {executed_query.url}"
        )
        if accept == "default":
            submitted_accept = executed_query.get_default_accept_header()
            app.echo_debug(
                f"Accept header set to default value: '{submitted_accept}'"
            )
        else:
            submitted_accept = accept

        results = executed_query.get_results(
            accept=submitted_accept,
            owl_imports_resolution=not no_imports,
            base64_encoded=base64,
            placeholder=placeholder,
            distinct=distinct,
            limit=limit,
            offset=offset,
            timeout=timeout
        )
        app.echo_result(results)


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "QUERIES",
    nargs=-1,
    required=True,
    autocompletion=completion.remote_queries_and_sparql_files
)
@click.pass_obj
def open_command(app, queries):
    """Open queries in the editor of the query catalog in your browser.

    With this command, you can open (remote) queries from the query catalog in
    the query editor in your browser (e.g. in order to change them).
    You can also load local query files into the query editor, in order to
    import them into the query catalog.

    The command accepts multiple query URIs or files which results in
    opening multiple browser tabs.
    """
    for file_or_uri in queries:
        opened_query = QUERY_CATALOG.get_query(file_or_uri)
        if opened_query is None:
            raise ValueError(
                f"{file_or_uri} is neither a (readable) file nor a query URI."
            )
        open_query_uri = opened_query.get_editor_url()
        app.echo_debug(f"Open {file_or_uri}: {open_query_uri}")
        click.launch(open_query_uri)


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only query identifier and no labels or other meta data. "
         "This is useful for piping the ids into other cmemc commands."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON response of the query status API."
)
@click.option(
    "--filter", "filter_",
    type=(str, str),
    multiple=True,
    autocompletion=completion.query_status_filter,
    help=QUERY_FILTER_HELP_TEXT,
)
@click.argument(
    "query_uuid",
    required=False,
    type=click.STRING
)
@click.pass_obj
def status_command(app, id_only, raw, filter_, query_uuid):
    """Get status information of executed and running queries.

    With this command, you can access the latest executed SPARQL queries
    on the DataPlatform. These queries are identified by UUIDs and listed
    ordered by starting timestamp.

    You can filter queries based on status and runtime in order to investigate
    slow queries. In addition to that, you can get the details of a specific
    query by using the ID as a parameter.
    """
    width, height = get_terminal_size((120, 20))
    max_query_string_width = width - 46 - 1
    app.echo_debug(f"Terminal size = {width} x {height}")

    queries = get_query_status()
    if query_uuid:
        queries = _get_queries_filtered(queries, "uuid", query_uuid)

    for _ in filter_:
        filter_type, filter_name = _
        queries = _get_queries_filtered(queries, filter_type, filter_name)
    if raw:
        app.echo_info_json(queries)
        return
    if id_only:
        for _ in queries:
            app.echo_info(_["id"])
        return
    if len(queries) == 1 and query_uuid:
        _output_query_status_details(app, queries[0])
        return
    table = []
    for _ in queries:
        query_string = " ".join(_["queryString"].splitlines())
        if len(query_string) > max_query_string_width:
            query_string = query_string[0:max_query_string_width] + "…"
        row = [
            _["id"],
            _["executionTime"],
            query_string
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=["Query ID", "Time", "Query String"]
    )


@click.command(cls=CmemcCommand, name="replay")
@click.argument(
    "REPLAY_FILE",
    required=True,
    autocompletion=completion.replay_files,
    type=click.Path(
        exists=True,
        allow_dash=False,
        readable=True,
        dir_okay=False
    )
)
@click.option(
    "--raw",
    is_flag=True,
    help="Output the execution statistic as raw JSON."
)
@click.option(
    "--loops",
    required=False,
    default=1,
    show_default=True,
    type=int,
    help="Number of loops to run the replay file."
)
@click.option(
    "--wait",
    required=False,
    default=0,
    show_default=True,
    type=int,
    help="Number of seconds to wait between query executions."
)
@click.option(
    "--output-file",
    required=False,
    autocompletion=completion.replay_files,
    help="Save the optional output to this file. Input and output of the "
         "command can be the same file. The output is written at the end "
         "of a successful command execution. The output can be stdout "
         "('-') - in this case, the execution statistic output is "
         "oppressed.",
    type=click.Path(
        exists=False,
        allow_dash=True,
        writable=True,
        dir_okay=False
    )
)
@click.option(
    "--run-label",
    type=click.STRING,
    help="Optional label of this replay run."
)
@click.pass_obj
def replay_command(
        app, replay_file, raw, loops, wait, output_file, run_label
):
    """Re-execute queries from a replay file.

    This command reads a REPLAY_FILE and re-executes the logged queries.
    A REPLAY_FILE is a JSON document which is an array of JSON objects with
    at least a key `queryString` holding the query text OR a key 'iri'
    holding the IRI of the query in the query catalog.
    It can be created with the `query status` command, e.g.
    `query status --raw > replay.json`

    The output of this command shows basic query execution statistics.

    The queries are executed one after another in the order given in the
    input REPLAY_FILE. Query placeholders / parameters are ignored. If a
    query results in an error, the duration is not counted.

    The optional output file is the same JSON document which is used as input,
    but each query object is annotated with an additional 'replays' object,
    which is an array of JSON objects which hold values for the
    replay|loop|run IDs, start and end time as well as duration and
    other data.
    """
    try:
        with open(replay_file, 'r', encoding='utf8') as _:
            input_queries = load(_)
    except JSONDecodeError as error:
        raise ValueError(
            f"File {replay_file} is not a valid JSON document."
        ) from error
    if len(input_queries) == 0:
        raise ValueError(
            f"File {replay_file} contains no queries."
        )
    app.echo_debug(
        f"File {replay_file} contains {len(input_queries)} queries."
    )

    statistic = ReplayStatistics(app=app, label=run_label)
    for _loop in range(loops):
        statistic.init_loop()
        for _ in input_queries:
            _ = statistic.measure_query_duration(_)
            if wait > 0:
                sleep(wait)

    if output_file:
        if output_file == "-":
            app.echo_info_json(input_queries)
            return
        with open(output_file, 'w', encoding='utf-8') as output:
            json.dump(input_queries, output, ensure_ascii=False, indent=2)

    if raw:
        statistic.output_json()
    else:
        statistic.output_table()


@click.group(cls=CmemcGroup)
def query():
    """List, execute, get status or open SPARQL queries.

    Queries are identified either by a file path, a URI from the query
    catalog or a shortened URI (qname, using a default namespace).

    In order to get a list of queries from the query catalog, use the list
    command. One or more queries can be executed one after the other with the
    execute command. With open command you can jump to the query editor in your
    browser.

    Queries can use a mustache like syntax to specify placeholder for
    parameter values (e.g. {{resourceUri}}). These parameter values need to
    be given as well, before the query can be executed (use the -p option).
    """


query.add_command(execute_command)
query.add_command(list_command)
query.add_command(open_command)
query.add_command(status_command)
query.add_command(replay_command)



"""cmemc commands."""
from click_help_colors import HelpColorsCommand, HelpColorsGroup

from click_didyoumean import DYMGroup

COLOR_FOR_COMMAND_GROUPS = "white"
COLOR_FOR_WRITING_COMMANDS = "red"
COLOR_FOR_HEADERS = "yellow"
COLOR_FOR_OPTIONS = "green"


# pylint: disable=too-many-ancestors
class CmemcGroup(HelpColorsGroup, DYMGroup):
    """Wrapper click.Group class to have a single extension point.

    Currently wrapped click extensions and additional group features:#
    - click-help-colors: https://github.com/click-contrib/click-help-colors
    """

    def __init__(self, *args, **kwargs):
        """Init a cmemc group command group."""
        # set default colors
        kwargs.setdefault("help_headers_color", COLOR_FOR_HEADERS)
        kwargs.setdefault("help_options_color", COLOR_FOR_OPTIONS)
        kwargs.setdefault(
            "help_options_custom_colors",
            {
                "bootstrap": COLOR_FOR_WRITING_COMMANDS,
                "showcase": COLOR_FOR_WRITING_COMMANDS,
                "delete": COLOR_FOR_WRITING_COMMANDS,
                "upload": COLOR_FOR_WRITING_COMMANDS,
                "import": COLOR_FOR_WRITING_COMMANDS,
                "create": COLOR_FOR_WRITING_COMMANDS,
                "enable": COLOR_FOR_WRITING_COMMANDS,
                "disable": COLOR_FOR_WRITING_COMMANDS,
                "execute": COLOR_FOR_WRITING_COMMANDS,
                "replay": COLOR_FOR_WRITING_COMMANDS,
                "io": COLOR_FOR_WRITING_COMMANDS,
                "install": COLOR_FOR_WRITING_COMMANDS,
                "uninstall": COLOR_FOR_WRITING_COMMANDS,
                "reload": COLOR_FOR_WRITING_COMMANDS,
                "update": COLOR_FOR_WRITING_COMMANDS,
                "eval": COLOR_FOR_WRITING_COMMANDS,
                "admin": COLOR_FOR_COMMAND_GROUPS,
                "store": COLOR_FOR_COMMAND_GROUPS,
                "metrics": COLOR_FOR_COMMAND_GROUPS,
                "config": COLOR_FOR_COMMAND_GROUPS,
                "dataset": COLOR_FOR_COMMAND_GROUPS,
                "graph": COLOR_FOR_COMMAND_GROUPS,
                "project": COLOR_FOR_COMMAND_GROUPS,
                "query": COLOR_FOR_COMMAND_GROUPS,
                "scheduler": COLOR_FOR_COMMAND_GROUPS,
                "vocabulary": COLOR_FOR_COMMAND_GROUPS,
                "workflow": COLOR_FOR_COMMAND_GROUPS,
                "workspace": COLOR_FOR_COMMAND_GROUPS,
                "python": COLOR_FOR_COMMAND_GROUPS,
                "cache": COLOR_FOR_COMMAND_GROUPS,
                "resource": COLOR_FOR_COMMAND_GROUPS,
            }
        )
        super().__init__(*args, **kwargs)


class CmemcCommand(HelpColorsCommand):
    """Wrapper click.Command class to have a single extension point.

    Currently wrapped click extensions and additional group features:#
    - click-help-colors: https://github.com/click-contrib/click-help-colors
    """

    def __init__(self, *args, **kwargs):
        """Init a cmemc group command.."""
        kwargs.setdefault('help_headers_color', COLOR_FOR_HEADERS)
        kwargs.setdefault('help_options_color', COLOR_FOR_OPTIONS)
        super().__init__(*args, **kwargs)



"""DataPlatform store commands for the cmem command line interface."""
import os
import sys
from os.path import exists

import click

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.context import ApplicationContext
from cmem.cmempy.dp.admin import create_showcase_data, import_bootstrap_data
from cmem.cmempy.dp.admin.backup import get_zip, post_zip


@click.command(cls=CmemcCommand, name="bootstrap")
@click.option(
    "--import", "import_",
    is_flag=True,
    help="Delete existing bootstrap data if present and import bootstrap "
         "data which was delivered "
)
@click.pass_obj
def bootstrap_command(app: ApplicationContext, import_):
    """Update/Import bootstrap data.

    This command imports the bootstrap data needed for managing shapes,
    access conditions, the query catalog and the vocabulary catalog.

    Note: There is currently no deletion mechanism for the bootstrap data,
    so you need to remove the graphs manually (or just remove all graphs).
    """
    if "store" not in sys.argv:
        app.echo_warning(
            "The 'admin bootstrap' command is deprecated and will be removed "
            "in the next major release.\n"
            "Please use the 'admin store bootstrap' command instead."
        )
    if not import_:
        raise ValueError("Use the --import flag to update/import "
                         "bootstrap data.")
    app.echo_info("Update or import bootstrap data ... ", nl=False)
    import_bootstrap_data()
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="showcase")
@click.option(
    "--scale",
    type=click.INT,
    default="10",
    show_default=True,
    help="The scale factor provides a way to set the target size of the "
         "scenario. A value of 10 results in around 40k triples, a value of "
         "50 in around 350k triples."
)
@click.option(
    "--create",
    is_flag=True,
    help="Delete old showcase data if present and create new showcase data"
         "based on the given scale factor."
)
@click.option(
    "--delete",
    is_flag=True,
    help="Delete existing showcase data if present."
)
@click.pass_obj
def showcase_command(app, scale, create, delete):
    """Create showcase data.

    This command creates a showcase scenario of multiple graphs including
    integration graphs, shapes, statement annotations etc.

    Note: There is currently no deletion mechanism for the showcase data, so
    you need to remove the showcase graphs manually (or just remove all
    graphs).
    """
    if "store" not in sys.argv:
        app.echo_warning(
            "The 'admin showcase' command is deprecated and will be removed "
            "in the next major release.\n"
            "Please use the 'admin store showcase' command instead."
        )
    if not delete and not create:
        raise ValueError("Either use the --create or the --delete flag.")
    if delete:
        raise NotImplementedError(
            "This feature is not implemented yet. "
            "Please delete the graphs manually."
        )
    if create:
        app.echo_info(
            f"Create showcase data with scale factor {scale} ... ",
            nl=False
        )
        create_showcase_data(scale_factor=scale)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="export")
@click.argument(
    "BACKUP_FILE",
    autocompletion=completion.graph_backup_files,
    required=True,
    type=click.Path(
        writable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrite existing files. "
         "This is a dangerous option, so use it with care.",
)
@click.pass_obj
def export_command(app, backup_file, overwrite):
    """Backup all knowledge graphs to a ZIP archive.

    The backup file is a ZIP archive containing all knowledge graphs as
    Turtle files + configuration file for each graph.

    This command will create lots of load on the server.
    It can take a long time to complete.
    """
    if exists(backup_file) and overwrite is not True:
        raise ValueError(
            f"Export file {backup_file} already exists and --overwrite "
            "option is not used."
        )
    with get_zip() as request:
        request.raise_for_status()
        with open(backup_file, 'wb') as _:
            requested_size = 1024 * 1024
            byte_counter = 0
            overall_byte_counter = 0
            app.echo_info(
                f"Exporting graphs backup to {backup_file} ...",
                nl=False
            )
            for chunk in request.iter_content(chunk_size=requested_size):
                chunk_size = len(chunk)
                app.echo_debug(f"Got new chuck of {chunk_size} bytes.")
                byte_counter += chunk_size
                overall_byte_counter += chunk_size
                _.write(chunk)
                _.flush()
                os.fsync(_.fileno())
                if byte_counter > requested_size:
                    app.echo_info(".", nl=False)
                    byte_counter = 0
            app.echo_debug(
                f"Wrote {overall_byte_counter} bytes to {backup_file}."
            )
            app.echo_success(" done")


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "BACKUP_FILE",
    autocompletion=completion.graph_backup_files,
    required=True,
    type=click.Path(
        readable=True,
        exists=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def import_command(app, backup_file):
    """Restore graphs from a ZIP archive.

    The backup file is a ZIP archive containing all knowledge graphs as
    Turtle files + configuration file for each graph.

    The command will load a single backup ZIP archive into the triple store,
    by replacing all graphs with the content of the Turtle files in the
    archive and deleting all graphs which are not in the archive.

    This command will create lots of load on the server.
    It can take a long time to complete.
    The backup file will be transferred to the server, then unzipped and
    imported graph by graph. After the initial transfer, the network
    connection is not used anymore, so it will be closed by proxies sometimes.
    This does not mean that the import failed.
    """
    app.echo_info(f"Importing graphs backup from {backup_file} ...", nl=False)
    request = post_zip(backup_file)
    request.raise_for_status()
    app.echo_success(" done")


@click.group(cls=CmemcGroup)
def store():
    """Import, export and bootstrap the knowledge graph store.

    This command group consist of commands to administrate the
    knowledge graph store as a whole.
    """


store.add_command(showcase_command)
store.add_command(bootstrap_command)
store.add_command(export_command)
store.add_command(import_command)



"""vocabularies commands for cmem command line interface."""
from datetime import date

from six.moves.urllib.parse import quote

import click

from rdflib import Graph
from rdflib.plugins.parsers.notation3 import BadSyntax

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.proxy import graph as graph_api
from cmem.cmempy.dp.titles import resolve
from cmem.cmempy.queries import SparqlQuery
from cmem.cmempy.vocabularies import (
    get_global_vocabs_cache,
    get_vocabularies,
    install_vocabulary,
    uninstall_vocabulary
)
from cmem.cmempy.workspace import (
    reload_prefixes,
    update_global_vocabulary_cache
)

GET_ONTOLOGY_IRI_QUERY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?iri
WHERE {
    ?iri a owl:Ontology;
}
"""

GET_PREFIX_DECLARATION = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX vann: <http://purl.org/vocab/vann/>
SELECT DISTINCT ?prefix ?namespace
WHERE {{
    <{}> a owl:Ontology;
        vann:preferredNamespacePrefix ?prefix;
        vann:preferredNamespaceUri ?namespace.
}}
"""

INSERT_CATALOG_ENTRY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX voaf: <http://purl.org/vocommons/voaf#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX dct: <http://purl.org/dc/terms/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX vann: <http://purl.org/vocab/vann/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
WITH <https://ns.eccenca.com/example/data/vocabs/>
INSERT {{
    <{iri}> a voaf:Vocabulary ;
        skos:prefLabel "{label}"{language} ;
        rdfs:label "{label}"{language} ;
        dct:description "{description}" ;
        vann:preferredNamespacePrefix "{prefix}" ;
        vann:preferredNamespaceUri "{namespace}" ;
        dct:modified "{date}"^^xsd:date .
}}
WHERE {{}}
"""


def _insert_catalog_entry(iri, prefix, namespace, label, language=None):
    """Insert a cmem vocabulary catalog entry.

    This executes an INSERT WHERE query to the vocabulary catalog graph in
    order to list the new vocabulary graph as vocab in the catalog.

    Args:
        iri (str): The IRI of the vocabulary graph.
        prefix (str): The prefix of the vocabulary.
        namespace (str): The namespace IRI of the vocabulary.
        label (str): The title of the vocabulary to add to the entry.
        language (str): Optional language tag of the title.

    Returns:
        None
    """
    if "@" + str(language).strip() != "@":
        language = "@" + str(language).strip()
    else:
        language = ""

    if not label.startswith(prefix + ":"):
        label = prefix + ": " + label

    query_text = INSERT_CATALOG_ENTRY.format(
        iri=iri,
        prefix=prefix,
        namespace=namespace,
        date=date.today().strftime("%Y-%m-%d"),
        label=label,
        language=language,
        description="vocabulary imported with cmemc"
    )
    query = SparqlQuery(text=query_text, origin="cmemc")
    query.get_results()


def _get_vocabulary_metadata_from_file(file):
    """Get potential graph iri and prefix/namespace from a turtle file."""
    metadata = {
        "iri": None,
        "prefix": None,
        "namespace": None
    }
    try:
        graph = Graph().parse(file, format="ttl")
    except BadSyntax as error:
        raise ValueError(
            "File {file} could not be parsed as turtle."
        ) from error

    ontology_iris = graph.query(GET_ONTOLOGY_IRI_QUERY)
    if len(ontology_iris) == 0:
        raise ValueError(
            "There is no owl:Ontology resource described "
            "in the turtle file."
        )
    if len(ontology_iris) > 1:
        ontology_iris_str = [str(iri[0]) for iri in ontology_iris]
        raise ValueError(
            "There are more than one owl:Ontology resources described "
            f"in the turtle file: {ontology_iris_str}"
        )
    iri = str(next(iter(ontology_iris))[0])
    metadata["iri"] = iri
    vann_data = graph.query(GET_PREFIX_DECLARATION.format(iri))
    if len(vann_data) == 0:
        raise ValueError(
            "There is no namespace defined "
            f"for the ontology '{iri}'.\n"
            "Please add a prefix and namespace to the sources.\n"
            "Refer to the documentation at "
            "https://vocab.org/vann/ for more information."
        )
    if len(vann_data) > 1:
        raise ValueError(
            "There is more than one vann namespace defined "
            f"for the ontology: {iri}"
        )
    namespace = next(iter(vann_data))
    metadata["prefix"] = str(namespace[0])
    metadata["namespace"] = str(namespace[1])
    return metadata


def _transform_cache_to_table(cache_category, table):
    """Transform a cache category dict to a tabulate table."""
    for item in cache_category:
        uri = item["genericInfo"]["uri"]
        try:
            label = item["genericInfo"]["label"]
        except KeyError:
            label = ""
        row = [uri, "class", label]
        table.append(row)
    return table


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "iri",
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.pass_obj
def open_command(app, iri):
    """Open / explore a vocabulary graph in the browser.

    Vocabularies are identified by their graph IRI.
    Installed vocabularies can be listed with the "vocabulary list" command.
    """
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary identifier (IRIs) and no labels or other "
         "meta data. This is useful for piping the ids into other cmemc "
         "commands."
)
@click.option(
    "--filter", "filter_",
    type=click.Choice(
        ["all", "installed", "installable"],
        case_sensitive=True
    ),
    default="installed",
    show_default=True,
    help="Filter list based on status."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def list_command(app, id_only, filter_, raw):
    """Output a list of vocabularies.

    Vocabularies are graphs (see 'cmemc graph' command group) which consists
    of class and property descriptions.
    """
    vocabs = get_vocabularies(filter_=filter_)
    if raw:
        app.echo_info_json(vocabs)
    elif id_only:
        for _ in vocabs:
            app.echo_info(_["iri"])
    else:
        table = []
        for _ in vocabs:
            iri = _["iri"]
            try:
                label = _["label"]["title"]
            except (KeyError, TypeError):
                if _["vocabularyLabel"]:
                    label = _["vocabularyLabel"]
                else:
                    label = "[no label given]"
            table.append((iri, label))
        app.echo_info_table(
            table,
            headers=["Vocabulary Graph IRI", "Label"],
            sort_column=1
        )


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installable_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Install all vocabularies from the catalog."
)
@click.pass_obj
def install_command(app, iris, all_):
    """Install one or more vocabularies from the catalog.

    Vocabularies are identified by their graph IRI.
    Installable vocabularies can be listed with the
    "vocabulary list --filter installable" command.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to install all "
                           "vocabularies from the catalog.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installable")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            f"Install vocabulary {current}/{count}: {iri} ... ",
            nl=False
        )
        install_vocabulary(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Uninstall all installed vocabularies."
)
@click.pass_obj
def uninstall_command(app, iris, all_):
    """Uninstall one or more vocabularies.

    Vocabularies are identified by their graph IRI.
    Already installed vocabularies can be listed with the
    "vocabulary list --filter installed" command.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to uninstall all "
                           "installed vocabularies.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installed")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            f"Uninstall vocabulary {current}/{count}: {iri} ... ",
            nl=False
        )
        uninstall_vocabulary(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "FILE",
    required=True,
    autocompletion=completion.triple_files,
    type=click.Path(
        allow_dash=False,
        readable=True
    )
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace (overwrite) existing vocabulary, if present."
)
@click.pass_obj
def import_command(app, file, replace):
    """Import a turtle file as a vocabulary.

    With this command, you can import a local ontology file as a named graph.
    and create a corresponding vocabulary catalog entry.

    The uploaded ontology file is analysed locally in order to discover the
    named graph and the prefix declaration. This requires an OWL ontology
    description which correctly uses the vann:preferredNamespacePrefix and
    vann:preferredNamespaceUri properties.
    """
    # fetch meta data
    meta_data = _get_vocabulary_metadata_from_file(file)
    iri = meta_data["iri"]
    prefix = meta_data["prefix"]
    namespace = meta_data["namespace"]

    success_message = "done"
    if iri in [_["iri"] for _ in graph_api.get_graphs_list()]:
        if replace:
            success_message = "replaced"
        else:
            raise ValueError(
                f"Proposed graph {iri} does already exist."
            )
    app.echo_info(
        f"Import {file} as vocabulary to {iri} ... ",
        nl=False
    )
    # upload graph
    graph_api.post(iri, file, replace=True)

    # resolve label
    resolved_label_object = resolve([iri], graph=iri)[iri]
    label = resolved_label_object["title"]
    language = resolved_label_object["lang"]
    app.echo_debug(resolved_label_object)

    # insert catalog entry
    _insert_catalog_entry(
        iri=iri,
        prefix=prefix,
        namespace=namespace,
        label=label,
        language=language
    )
    # reload DI prefix
    reload_prefixes()
    # update cache
    update_global_vocabulary_cache(iri)
    app.echo_success(success_message)


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.installed_vocabularies
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Update cache for all installed vocabularies."
)
@click.pass_obj
def cache_update_command(app, iris, all_):
    """Reload / updates the data integration cache for a vocabulary."""
    if iris == () and not all_:
        raise ValueError("Either specify at least one vocabulary "
                         + "IRI or use the --all option to update the "
                           "cache for all installed vocabularies.")
    if all_:
        iris = (
            [iri["iri"] for iri in get_vocabularies(filter_="installed")]
        )
    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info(
            f"Update cache {current}/{count}: {iri} ... ",
            nl=False
        )
        update_global_vocabulary_cache(iri)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary term identifier (IRIs) and no labels or other "
         "meta data. This is useful for piping the ids into other cmemc "
         "commands."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def cache_list_command(app, id_only, raw):
    """Output the content of the global vocabulary cache."""
    cache_ = get_global_vocabs_cache()
    if raw:
        app.echo_info_json(cache_)
    elif id_only:
        for vocab in cache_["vocabularies"]:
            for class_ in vocab["classes"]:
                app.echo_info(class_["genericInfo"]["uri"])
            for property_ in vocab["properties"]:
                app.echo_info(property_["genericInfo"]["uri"])
    else:
        table = []
        for vocab in cache_["vocabularies"]:
            table = _transform_cache_to_table(vocab["classes"], table)
            table = _transform_cache_to_table(vocab["properties"], table)
        app.echo_info_table(
            table,
            headers=["IRI", "Type", "Label"],
            sort_column=0
        )


@click.group(cls=CmemcGroup)
def cache():
    """List und update the vocabulary cache."""


cache.add_command(cache_update_command)
cache.add_command(cache_list_command)


@click.group(cls=CmemcGroup)
def vocabulary():
    """List, (un-)install, import or open vocabs / manage cache."""


vocabulary.add_command(open_command)
vocabulary.add_command(list_command)
vocabulary.add_command(install_command)
vocabulary.add_command(uninstall_command)
vocabulary.add_command(import_command)
vocabulary.add_command(cache)



"""The main command line interface."""

import os
from subprocess import CalledProcessError  # nosec
import sys
import traceback

import click

import requests.exceptions

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import (
    admin,
    config,
    dataset,
    graph,
    project,
    query,
    vocabulary,
    workflow
)
from cmem.cmemc.cli.context import CONTEXT
from cmem.cmemc.cli.exceptions import (
    InvalidConfiguration
)
from cmem.cmemc.cli.utils import (
    extract_error_message,
    get_version,
    is_completing,
    print_manual,
    print_manual_graph
)
from cmem.cmemc.cli.commands import CmemcGroup

CMEMC_VERSION = get_version()

version = sys.version_info
PYTHON_VERSION = f"{version.major}.{version.minor}.{version.micro}"
PYTHON_EXPECTED = "3.9"
PYTHON_GOT = f"{version.major}.{version.minor}"
if PYTHON_EXPECTED != PYTHON_GOT:
    # test for environment which indicates that we are in completion mode
    if not is_completing():
        CONTEXT.echo_warning(
            "Warning: You are running cmemc under a non-tested python "
            f"environment (expected {PYTHON_EXPECTED}, got {PYTHON_GOT})"
        )

# set the user-agent environment for the http request headers
os.environ["CMEM_USER_AGENT"] = f"cmemc/{CMEMC_VERSION} " \
                                f"(Python {PYTHON_VERSION})"

# https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
CONTEXT_SETTINGS = dict(
    auto_envvar_prefix='CMEMC',
    help_option_names=['-h', '--help']
)


@click.group(cls=CmemcGroup, context_settings=CONTEXT_SETTINGS)
@click.option(
    '-c', '--connection',
    type=click.STRING,
    autocompletion=completion.connections,
    help='Use a specific connection from the config file.'
)
@click.option(
    '--config-file',
    autocompletion=completion.ini_files,
    type=click.Path(
        readable=True,
        allow_dash=False,
        dir_okay=False
    ),
    default=CONTEXT.config_file_default,
    show_default=True,
    help='Use this config file instead of the default one.'
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help='Suppress any non-error info messages.'
)
@click.option(
    '-d', '--debug',
    is_flag=True,
    help='Output debug messages and stack traces after errors.'
)
@click.version_option(
    version=CMEMC_VERSION,
    message="%(prog)s, version %(version)s, "
            f"running under python {PYTHON_VERSION}"
)
@click.pass_context
def cli(ctx, debug, quiet, config_file, connection):  # noqa: D403
    """eccenca Corporate Memory Control (cmemc).

    cmemc is the eccenca Corporate Memory Command Line Interface (CLI).

    Available commands are grouped by affecting resource type (such as graph,
    project and query).
    Each command and group has a separate --help screen for detailed
    documentation.
    In order to see possible commands in a group, simply
    execute the group command without further parameter (e.g. cmemc project).

    If your terminal supports colors, these coloring rules are applied:
    Groups are colored in white; Commands which change data are colored in
    red; all other commands as well as options are colored in green.

    Please also have a look at the cmemc online documentation:

                        https://eccenca.com/go/cmemc

    cmemc is © 2022 eccenca GmbH, licensed under the Apache License 2.0.
    """
    ctx.obj = CONTEXT
    # hidden feature: 'CMEMC_MANUAL=true cmemc -q config list' will output
    #     the whole markdown manual
    if os.getenv("CMEMC_MANUAL"):
        print_manual(ctx)
        ctx.exit()
    # hidden feature: 'CMEMC_MANUAL_GRAPH=true cmemc -q config list' will
    # output the documentation graph
    if os.getenv("CMEMC_MANUAL_GRAPH"):
        print_manual_graph(ctx)
        ctx.exit()
    ctx.obj.set_quiet(quiet)
    ctx.obj.set_debug(debug)
    ctx.obj.set_config_file(config_file)
    try:
        ctx.obj.set_connection(connection)
    except InvalidConfiguration as error:
        # if config is broken still allow for "config edit"
        # means: do not forward this exception if "config edit"
        if " ".join(sys.argv).find("config edit") == -1:
            raise error


cli.add_command(admin.admin)
cli.add_command(config.config)
cli.add_command(dataset.dataset)
cli.add_command(graph.graph)
cli.add_command(project.project)
cli.add_command(query.query)
cli.add_command(vocabulary.vocabulary)
cli.add_command(workflow.workflow)


def main():
    """Start the command line interface."""
    try:
        cli()  # pylint: disable=no-value-for-parameter
    except (
            CalledProcessError,
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            ValueError,
            IOError,
            NotImplementedError,
            KeyError
    ) as error:
        if is_completing():
            # if currently autocompleting -> silently die with exit 1
            sys.exit(1)
        CONTEXT.check_versions()
        CONTEXT.echo_debug(traceback.format_exc())
        CONTEXT.echo_error(extract_error_message(error))
        sys.exit(1)



"""DataIntegration dataset resource commands for cmemc."""
import re

import click

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.utils import split_task_id, struct_to_table
from cmem.cmempy.workspace.projects.resources import get_all_resources
from cmem.cmempy.workspace.projects.resources.resource import (
    delete_resource,
    get_resource_metadata,
    get_resource_usage_data
)

RESOURCE_FILTER_TYPES = ["project", 'regex']
RESOURCE_FILTER_TYPES_HIDDEN = ["ids"]
RESOURCE_FILTER_TEXT = (
    "Filter file resources based on a meta data. "
    f"First parameter CHOICE can be one of {str(RESOURCE_FILTER_TYPES)}"
    ". The second parameter is based on CHOICE, e.g. a project "
    "ID or a regular expression string."
)


def _get_resources_filtered(resources, filter_name, filter_value):
    """Get file resources but filtered according to name and value."""
    filter_types = RESOURCE_FILTER_TYPES + RESOURCE_FILTER_TYPES_HIDDEN
    # check for correct filter names (filter ids is used internally only)
    if filter_name not in filter_types:
        raise ValueError(
            f"{filter_name} is an unknown filter name. "
            f"Use one of {RESOURCE_FILTER_TYPES}."
        )
    # filter by ID list
    if filter_name == "ids":
        return [_ for _ in resources if _["id"] in filter_value]
    # filter by project
    if filter_name == "project":
        return [_ for _ in resources if _["project"] == filter_value]
    # filter by regex
    if filter_name == "regex":
        return [_ for _ in resources if re.search(filter_value, _["name"])]
    # return unfiltered list
    return resources


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only resource names and no other meta data. "
         "This is useful for piping the IDs into other commands."
)
@click.option(
    "--filter", "filters_",
    multiple=True,
    type=(str, str),
    autocompletion=completion.resource_list_filter,
    help=RESOURCE_FILTER_TEXT
)
@click.pass_obj
def list_command(app, raw, id_only, filters_):
    """List available file resources.

    Outputs a table or a list of dataset resources (files).
    """
    resources = get_all_resources()
    for _ in filters_:
        filter_name, filter_value = _
        resources = _get_resources_filtered(
            resources, filter_name, filter_value
        )
    if raw:
        app.echo_info_json(resources)
        return
    if id_only:
        for _ in sorted(_["id"] for _ in resources):
            app.echo_result(_)
        return
    # output a user table
    table = []
    headers = ["ID", "Modified", "Size"]
    for _ in resources:
        row = [
            _["id"],
            _["modified"],
            _["size"],
        ]
        table.append(row)
    app.echo_info_table(table, headers=headers, sort_column=0)


@click.command(cls=CmemcCommand, name="delete")
@click.argument(
    "resource_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.resource_ids
)
@click.option(
    "--force",
    is_flag=True,
    help="Delete resource even if in use by a task."
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Delete all resources. "
         "This is a dangerous option, so use it with care.",
)
@click.option(
    "--filter", "filters_",
    multiple=True,
    type=(str, str),
    autocompletion=completion.resource_list_filter,
    help=RESOURCE_FILTER_TEXT
)
@click.pass_obj
def delete_command(app, resource_ids, force, all_, filters_):
    """Delete file resources.

    You have three selection mechanisms: with specific IDs, you will delete
    only these resources; by using --filter your will delete resources based
    on the filter type and value; by using --all will delete all resources.
    """
    if resource_ids == () and not all_ and filters_ == ():
        raise ValueError(
            "Either specify at least one resource ID or use the --all or "
            "--filter options to specify resources for deletion."
        )

    resources = get_all_resources()
    if len(resource_ids) > 0:
        for resource_id in resource_ids:
            if resource_id not in [_["id"] for _ in resources]:
                raise ValueError(
                    f"Resource {resource_id} not available."
                )
        # "filter" by id
        resources = _get_resources_filtered(
            resources, "ids", resource_ids
        )
    for _ in filters_:
        resources = _get_resources_filtered(resources, _[0], _[1])

    # avoid double removal as well as sort IDs
    resource_ids = sorted(
        list({_["id"] for _ in resources}),
        key=lambda v: v.lower()
    )
    count = len(resource_ids)
    length = len(str(count))
    current = 0
    for resource_id in resource_ids:
        current = current + 1
        current_string = str(current).zfill(length)
        app.echo_info(
            f"Delete resource {current_string}/{count}: {resource_id} ... ",
            nl=False
        )
        project_id, resource_local_id = split_task_id(resource_id)
        usage = get_resource_usage_data(project_id, resource_local_id)
        if len(usage) > 0:
            app.echo_error(f"in use by {len(usage)} task(s)", nl=False)
            if force:
                app.echo_info(" ... ", nl=False)
            else:
                app.echo_info("")
                continue
        delete_resource(
            project_name=project_id, resource_name=resource_local_id
        )
        app.echo_success("deleted")


@click.command(cls=CmemcCommand, name="inspect")
@click.argument(
    "resource_id",
    type=click.STRING,
    autocompletion=completion.resource_ids
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def inspect_command(app, resource_id, raw):
    """Display all meta data of a file resource."""
    project_id, resource_id = split_task_id(resource_id)
    resource_data = get_resource_metadata(project_id, resource_id)
    if raw:
        app.echo_info_json(resource_data)
    else:
        table = struct_to_table(resource_data)
        app.echo_info_table(
            table,
            headers=["Key", "Value"],
            sort_column=0
        )


@click.command(cls=CmemcCommand, name="usage")
@click.argument(
    "resource_id",
    type=click.STRING,
    autocompletion=completion.resource_ids
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.pass_obj
def usage_command(app, resource_id, raw):
    """Display all usage data of a file resource."""
    project_id, resource_id = split_task_id(resource_id)
    usage = get_resource_usage_data(project_id, resource_id)
    if raw:
        app.echo_info_json(usage)
        return
    # output a user table
    table = []
    headers = ["Task ID", "Type", "Label"]
    for _ in usage:
        row = [
            project_id + ":" + _["id"],
            _["taskType"],
            _["label"]
        ]
        table.append(row)
    app.echo_info_table(table, headers=headers, sort_column=2)


@click.group(cls=CmemcGroup)
def resource():
    """List, inspect or delete dataset file resources.

    File resources are identified by its name and project ID.
    """


resource.add_command(list_command)
resource.add_command(delete_command)
resource.add_command(inspect_command)
resource.add_command(usage_command)



"""workflow commands for cmem command line interface."""
import datetime
import os
import sys
import time

import click

import requests

import timeago

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.commands.scheduler import scheduler
from cmem.cmempy.workflow import get_workflows
from cmem.cmempy.workflow.workflow import (
    execute_workflow_io,
    get_workflow_editor_uri,
    get_workflows_io
)
from cmem.cmempy.workspace.activities import (
    ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW,
    VALID_ACTIVITY_STATUS
)
from cmem.cmempy.workspace.activities.taskactivities import (
    get_activities_status
)
from cmem.cmempy.workspace.activities.taskactivity import (
    get_activity_status,
    start_task_activity
)
from cmem.cmempy.workspace.projects.project import get_projects
from cmem.cmempy.workspace.search import list_items

IO_WARNING_NO_RESULT = "The workflow was executed but produced no result."
IO_WARNING_NO_OUTPUT_DEFINED = "The workflow was executed, a result was "\
                               "received but dropped."

MIME_CSV = "text/csv"
MIME_XLS = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
MIME_NT = "application/n-triples"
MIME_JSON = "application/json"
MIME_XML = "application/xml"

VALID_OUTPUT_EXTENSIONS = {
    ".csv": MIME_CSV,
    ".xlsx": MIME_XLS,
    ".nt": MIME_NT,
    ".ttl": MIME_NT,
    ".json": MIME_JSON,
    ".xml": MIME_XML
}

VALID_INPUT_EXTENSIONS = {
    ".csv": MIME_CSV,
    ".json": MIME_JSON,
    ".xml": MIME_XML
}


def _get_workflows_filtered_by_project(project):
    """Get workflows but filtered by project.

    Args:
        project (str): project ID

    Returns:
        list of filtered workflows form the list_items API call
    """
    return list_items(
        project=project, item_type="workflow"
    )["results"]


def _get_workflows_filtered_by_io_feature(feature):
    """Get workflows filtered by io feature.

    Args:
        feature (str): one of input-only|output-only|input-output or any

    Returns:
        list of filtered workflows form the list_items API call

    Raises:
        ValueError
    """
    possible_io_filter_values = (
        "input-only", "output-only", "input-output", "any"
    )
    if feature not in possible_io_filter_values:
        raise ValueError(
            f"{feature} is an unknown filter value. "
            f"Use one of {possible_io_filter_values}."
        )
    filtered_workflows_ids = []
    for _ in get_workflows_io():
        ins = len(_["variableInputs"])
        outs = len(_["variableOutputs"])
        if feature == "any" and (ins == 1 or outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "input-only" and (ins == 1 and outs == 0):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "output-only" and (ins == 0 and outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "input-output" and (ins == 1 and outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
    filtered_workflows = []
    for _ in list_items(item_type="workflow")["results"]:
        if _["projectId"] + ":" + _["id"] in filtered_workflows_ids:
            filtered_workflows.append(_)
    return filtered_workflows


def _get_workflows_filtered(filter_name, filter_value):
    """Get workflows filtered according to filter name and value.

    Args:
        filter_name (str): one of "project" or "io"
        filter_value (str): value according to fileter

    Returns:
        list of filtered workflows from the list_items API call

    Raises:
        ValueError
    """
    # filter by project
    if filter_name == "project":
        return _get_workflows_filtered_by_project(filter_value)
    # filter by io feature
    if filter_name == "io":
        return _get_workflows_filtered_by_io_feature(filter_value)
    # default is unfiltered
    return list_items(item_type="workflow")["results"]


def _io_check_request(
        info=None,
        input_file=None,
        output_file=None,
        output_mimetype=None
):
    """Check the requested io execution."""
    if len(info["variableInputs"]) == 0 and input_file:
        raise ValueError(
            "You are trying to send data to a workflow without a variable "
            "input. Please remove the '-i' parameter."
        )
    if len(info["variableOutputs"]) == 0 and output_file:
        raise ValueError(
            "You are trying to retrieve data to a workflow without a variable "
            "output. Please remove the '-o' parameter."
        )
    if len(info["variableInputs"]) == 1 and not input_file:
        raise ValueError(
            "This workflow has a defined input so you need to use the '-i' "
            "parameter to send data to it."
        )
    if len(info["variableOutputs"]) == 1 and not output_file:
        raise ValueError(
            "This workflow has a defined output so you need to use the '-o' "
            "parameter to retrieve data from it."
        )
    if output_mimetype == MIME_XLS and output_file == "-":
        raise ValueError(
            "Trying to output an excel document to stdout will fail.\n"
            "Please output to a regular file instead "
            "(workflow was not executed)."
        )


def _io_get_info(project_id, workflow_id):
    """Get the io info dictionary of the workflow."""
    for _ in get_workflows_io():
        if _["id"] == workflow_id and _["projectId"] == project_id:
            return _
    raise ValueError(
        "The given workflow does not exist or is not suitable to be executed "
        "with this command.\n"
        "An io workflow needs exactly one variable input and/or one variable "
        "output."
    )


def _io_process_response(response, app, output_file):
    """Process the workflow response of the io command."""
    with response:
        if output_file == "-":
            for line in response.iter_lines():
                if line:
                    app.echo_info(line.decode("UTF-8"))
        else:
            with click.open_file(output_file, 'wb') as file:
                for chunk in response.iter_content(chunk_size=8192):
                    file.write(chunk)


def _io_guess_output(output_file):
    """Guess the mime type of an output file name."""
    if output_file == "-":
        return None
    _, file_extension = os.path.splitext(output_file)
    if file_extension in VALID_OUTPUT_EXTENSIONS:
        return VALID_OUTPUT_EXTENSIONS[file_extension]
    valid_extensions = ', '.join(VALID_OUTPUT_EXTENSIONS.keys())
    raise ValueError(
        f"Files with the extension {file_extension} can not be generated. "
        f"Try one of {valid_extensions}"
    )


def _io_guess_input(input_file):
    """Guess the mime type of an input file name."""
    if input_file == "-":
        return None
    _, file_extension = os.path.splitext(input_file)
    if file_extension in VALID_INPUT_EXTENSIONS:
        return VALID_INPUT_EXTENSIONS[file_extension]
    valid_extensions = ', '.join(VALID_INPUT_EXTENSIONS.keys())
    raise ValueError(
        f"Files with the extension {file_extension} can not be processed. "
        f"Try one of {valid_extensions}"
    )


def _workflows_get_ids():
    """Get a list of workflow IDs."""
    ids = []
    for project_desc in get_projects():
        for workflow_id in get_workflows(project_desc["name"]):
            ids.append(project_desc["name"] + ":" + workflow_id)
    return ids


def _workflow_wait_until_finished(project_id, task_id, polling_interval):
    """Poll workflow status until workflow is finished and return status.

    :type project_id: string
    :type task_id: string
    :type polling_interval: int
    """
    while True:
        status = get_activity_status(project_id, task_id)
        # wait until isRunning is false
        if not status['isRunning']:
            break
        time.sleep(polling_interval)
    return status


def _status_command_pre_cmem_20_10(
        app, raw, _filter, workflow_ids
):
    """Get status of workflows without new taskActivitiesStatus API."""
    if workflow_ids == ():
        workflow_ids += tuple(_workflows_get_ids())

    for workflow_id in workflow_ids:
        project_id, task_id = workflow_id.split(":")
        status = get_activity_status(project_id, task_id)
        # filter for running activities
        if _filter == "running"\
                and 'isRunning' in status and not status['isRunning']:
            continue
        # filter for failed activities
        if _filter == "failed"\
                and "failed" in status and not status['failed']:
            continue
        # output raw json if wanted
        if raw:
            app.echo_info_json(status)
        # output single line status
        else:
            app.echo_info(f"{workflow_id} ... ", nl=False)
            _workflow_echo_status(app, status)


def _workflow_echo_status(app, status):
    """Print a colored status based on the returned JSON.

    Status can be Idle, Running, Canceling, Waiting, Finished
    isRunning is true for Running, Canceling, Waiting
    canceled only exists sometimes
    """
    # prepare human friendly relative time
    now = datetime.datetime.now()
    stamp = datetime.datetime.fromtimestamp(status["lastUpdateTime"] / 1000)
    time_ago = timeago.format(stamp, now, 'en')
    status_name = status["statusName"]
    status_message = status["message"]
    # prepare message
    if status_name == status["message"]:
        message = f"{status_name} ({time_ago})"
    else:
        message = f"{status_name} ({status_message}, {time_ago})"

    if status["isRunning"]:
        if status_name == "Running":
            app.echo_warning(message)
        elif status_name == "Canceling":
            app.echo_warning(message)
        elif status_name == "Waiting":
            app.echo_warning(message)
        else:
            raise ValueError(
                f"statusName is {status_name}, expecting one of: "
                "Running, Canceling or Waiting."
            )
    else:
        # not running can be Idle or Finished
        if "failed" in status and status["failed"]:
            app.echo_error(message, nl=True, err=False)
        elif "cancelled" in status and status["cancelled"]:
            app.echo_warning(message)
        elif status["statusName"] == "Idle":
            app.echo_info(message)
        else:
            # Finished and without failed or canceled status
            app.echo_success(message)


@click.command(cls=CmemcCommand, name="execute")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Execute all available workflows."
)
@click.option(
    "--wait",
    is_flag=True,
    help="Wait until all executed workflows are completed."
)
@click.option(
    "--polling-interval",
    type=click.IntRange(min=0, max=60),
    show_default=True,
    default=1,
    help="How many seconds to wait between status polls. Status polls are"
         " cheap, so a higher polling interval is most likely not needed."
)
@click.argument(
    "workflow_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def execute_command(app, all_, wait, polling_interval, workflow_ids):
    """Execute workflow(s).

    With this command, you can start one or more workflows at the same time or
    in a sequence, depending on the result of the predecessor.

    Executing a workflow can be done in two ways: Without --wait just sends
    the starting signal and does not look for the workflow and its result
    (fire and forget). Starting workflows in this way, starts all given
    workflows at the same time.

    The optional --wait option starts the workflows in the same way, but also
    polls the status of a workflow until it is finished. In case of an error of
    a workflow, the next workflow is not started.
    """
    all_workflow_ids = _workflows_get_ids()
    if workflow_ids == () and not all_:
        raise ValueError("Either specify at least one workflow or use the"
                         + " --all option to execute all workflows.")
    if all_:
        workflow_ids = all_workflow_ids

    for workflow_id in workflow_ids:
        if workflow_id not in all_workflow_ids:
            raise ValueError(
                f"Workflow '{workflow_id}' is not available."
            )
        project_id, task_id = workflow_id.split(":")
        app.echo_info(f"{workflow_id} ... ", nl=False)

        # before we start, we fetch the status
        status = get_activity_status(project_id, task_id)
        if not wait:
            if status['isRunning']:
                # in case of a running workflow, we only output status
                app.echo_info('Already Running')
            else:
                # in case of simple call, we just start and forget
                start_task_activity(project_id, task_id)
                app.echo_info('Started')
        else:
            # in case of --wait, we poll the status until finished
            if status['isRunning']:
                # in case of a running workflow, we only output status
                app.echo_info('Already Running ... ', nl=False)
            else:
                start_task_activity(project_id, task_id)
                app.echo_info('Started ... ', nl=False)

            status = _workflow_wait_until_finished(
                project_id, task_id, polling_interval
            )
            # when we have a Finished status, we output it
            _workflow_echo_status(app, status)
            # in case of failure, the following workflows are not executed
            if status['failed']:
                sys.exit(1)


# pylint: disable=too-many-arguments
@click.command(cls=CmemcCommand, name="io")
@click.option(
    "--input", "-i", "input_file",
    type=click.Path(
        allow_dash=False,
        dir_okay=False,
        readable=True
    ),
    autocompletion=completion.workflow_io_input_files,
    help="From which file the input is taken: note that the maximum file size "
         "to upload is limited to a server configured value. If the workflow "
         "has no defined variable input dataset, this can be ignored."
)
@click.option(
    "--output", "-o", "output_file",
    type=click.Path(
        allow_dash=False,
        dir_okay=False,
        writable=True,
    ),
    autocompletion=completion.workflow_io_output_files,
    help="To which file the result is written to: use '-' in order to output "
         "the result to stdout. If the workflow has no defined variable "
         "output dataset, this can be ignored. Please note that the io "
         "command will not warn you on overwriting existing output files."
)
@click.option(
    "--input-mimetype",
    help="Which input format should be processed: If not given, cmemc will "
         "try to guess the mime type based on the file extension or will "
         "fail",
    type=click.Choice(
        [
            "guess",
            "application/xml",
            "application/json",
            "text/csv"
        ]
    ),
    default="guess"
)
@click.option(
    "--output-mimetype",
    help="Which output format should be requested: If not given, cmemc will "
         "try to guess the mime type based on the file extension or will "
         "fail. In case of an output to stdout, a default mime type "
         "will be used (currently xml).",
    type=click.Choice(
        ["guess", MIME_XML, MIME_JSON, MIME_NT, MIME_XLS, MIME_CSV]
    ),
    default="guess"
)
@click.argument(
    "workflow_id",
    type=click.STRING,
    autocompletion=completion.workflow_io_ids
)
@click.pass_obj
def io_command(
        app,
        input_file, input_mimetype,
        output_file, output_mimetype,
        workflow_id
):
    """Execute a workflow with file input/output.

    With this command, you can execute a workflow that uses variable datasets
    as input, output or for configuration. Use the input parameter to feed
    data into the workflow. Likewise use output for retrieval of the workflow
    result. Workflows without a variable dataset will throw an error.
    """
    project_id, task_id = workflow_id.split(":")
    if output_file and output_mimetype == "guess":
        output_mimetype = _io_guess_output(output_file)
    if input_file and input_mimetype == "guess":
        input_mimetype = _io_guess_input(input_file)

    _io_check_request(
        info=_io_get_info(project_id, task_id),
        input_file=input_file,
        output_file=output_file,
        output_mimetype=output_mimetype
    )

    response = execute_workflow_io(
        project_name=project_id,
        task_name=task_id,
        input_file=input_file,
        input_mime_type=input_mimetype,
        output_mime_type=output_mimetype
    )
    app.echo_debug(
        f"Workflow response received after {response.elapsed} "
        f"with status {response.status_code}."
    )
    if response.status_code == 204:
        # empty content, warn if output was requested
        # this will happen only if info was wrong
        if output_file:
            app.echo_warning(IO_WARNING_NO_RESULT)
        # execution successful
        return
    if response.status_code == 200 and not output_file:
        # returns with content, warn if NO output was requested
        # this will happen only if info was wrong
        app.echo_warning(IO_WARNING_NO_OUTPUT_DEFINED)
        # execution successful
        return
    _io_process_response(response, app, output_file)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON objects of workflow task search API response."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only workflow identifier and no labels or other "
         "meta data. This is useful for piping the IDs into other commands."
)
@click.option(
    "--filter", "filter_",
    type=click.Tuple([
        click.Choice(["project", 'io']),
        str
    ]),
    autocompletion=completion.workflow_list_filter,
    default=[None] * 2,
    help="Filter workflows based on project or suitability for the io "
         "command ."
         "First parameter CHOICE can be 'project' or 'io'. "
         "The second parameter has to be a project ID in case "
         "of 'project' or 'input-only|output-only|input-output|any' in case "
         "of 'io' filter."
)
@click.pass_obj
def list_command(app, raw, id_only, filter_):
    """List available workflow ids."""
    filter_name, filter_value = filter_
    workflows = _get_workflows_filtered(filter_name, filter_value)
    if raw:
        app.echo_info_json(workflows)
        return
    if id_only:
        # sort by combined project + task ID
        for _ in sorted(
                workflows,
                key=lambda k: (k["projectId"] + ":" + k["id"]).lower()
        ):
            app.echo_info(_["projectId"] + ":" + _["id"])
        return
    # output a user table
    table = []
    for _ in workflows:
        row = [
            _["projectId"] + ":" + _["id"],
            _["label"],
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=["Workflow ID", "Label"],
        sort_column=1
    )


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--project", "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids,
    help="The project, from which you want to list the workflows. "
         "Project IDs can be listed with the 'project list' command."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Output raw JSON info."
)
@click.option(
    "--filter", "_filter",
    type=click.Choice(VALID_ACTIVITY_STATUS, case_sensitive=False),
    help="Show only workflows of a specific status."
)
@click.argument(
    "workflow_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def status_command(app, project_id, raw, _filter, workflow_ids):
    """Get status information of workflow(s)."""
    try:
        workflow_status = get_activities_status(
            status=_filter,
            project_name=project_id,
            activity=ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW
        )
        for status in workflow_status:
            workflow_id = status["project"] + ":" + status["task"]
            if len(workflow_ids) == 0 or workflow_id in workflow_ids:
                if raw:
                    # TODO: output for more than one workflow is not valid JSON
                    app.echo_info_json(status)
                else:
                    app.echo_info(f"{workflow_id} ... ", nl=False)
                    _workflow_echo_status(app, status)
    except requests.exceptions.HTTPError:
        app.echo_warning("CMEM 20.10 taskActivitiesStatus API not available.\n"
                         "Falling back to older and more slow API with less "
                         "filter options.")
        _status_command_pre_cmem_20_10(app, raw, _filter, workflow_ids)


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "workflow_id",
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def open_command(app, workflow_id):
    """Open a workflow in your browser."""
    project_id, task_id = workflow_id.split(":")
    workflow_editor_uri = get_workflow_editor_uri()\
        .format(project_id, task_id)
    click.launch(workflow_editor_uri)
    app.echo_debug(workflow_editor_uri)


@click.group(cls=CmemcGroup)
def workflow():
    """List, execute, status or open (io) workflows.

    Workflows are identified by a WORKFLOW_ID. The get a list of existing
    workflows, execute the list command or use tab-completion.
    The WORKFLOW_ID is a concatenation of an PROJECT_ID and a TASK_ID, such as
    "my-project:my-workflow".
    """


workflow.add_command(execute_command)
workflow.add_command(io_command)
workflow.add_command(list_command)
workflow.add_command(status_command)
workflow.add_command(open_command)
workflow.add_command(scheduler)



"""workspace commands for cmem command line interface."""
import os
from pathlib import Path

import click

from jinja2 import Template

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.commands.python import python
from cmem.cmempy.workspace import reload_workspace
from cmem.cmempy.workspace.export_ import export
from cmem.cmempy.workspace.import_ import import_workspace


@click.command(cls=CmemcCommand, name="export")
@click.option(
    "-o", "--overwrite",
    is_flag=True,
    help="Overwrite existing files. "
         "This is a dangerous option, so use it with care.",
)
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported workspace file."
)
@click.option(
    "--filename-template", "-t", "template",
    default="{{date}}-{{connection}}.workspace",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.workspace_export_templates,
    help="Template for the export file name. "
         "Possible placeholders are (Jinja2): "
         "{{connection}} (from the --connection option) and "
         "{{date}} (the current date as YYYY-MM-DD). "
         "The file suffix will be appended. "
         "Needed directories will be created."
)
@click.argument(
    "file",
    autocompletion=completion.workspace_files,
    required=False,
    type=click.Path(
        writable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def export_command(app, overwrite, marshalling_plugin, template, file):
    """Export the complete workspace (all projects) to a ZIP file.

    Depending on the requested type, this ZIP contains either a turtle file
    for each project (type rdfTurtle) or a substructure of resource files and
    XML descriptions (type xmlZip).

    The file name is optional and will be generated with by
    the template if absent.
    """
    ctx = click.get_current_context()
    if ctx.auto_envvar_prefix == "CMEMC_WORKSPACE_EXPORT":
        app.echo_warning(
            "The command workspace export is deprecated. "
            "Please use 'cmemc admin workspace export' instead."
        )
    if file is None:
        # prepare the template data and create the actual file incl. suffix
        template_data = app.get_template_data()
        file = Template(template).render(template_data) + ".zip"
    file = os.path.normpath(file)
    app.echo_info(f"Export workspace to {file} ... ", nl=False)
    if os.path.exists(file) and overwrite is not True:
        app.echo_error("file exists")
    else:
        # output directory is created lazy
        Path(
            os.path.dirname(file)
        ).mkdir(parents=True, exist_ok=True)
        # do the export
        export_data = export(marshalling_plugin)
        with open(file, "wb") as export_file:
            export_file.write(export_data)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="import")
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported workspace file."
)
@click.argument(
    "file",
    autocompletion=completion.workspace_files,
    type=click.Path(
        readable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def import_command(app, file, marshalling_plugin):
    """Import the workspace from a file."""
    ctx = click.get_current_context()
    if ctx.auto_envvar_prefix == "CMEMC_WORKSPACE_IMPORT":
        app.echo_warning(
            "The command workspace import is deprecated. "
            "Please use 'cmemc admin workspace import' instead."
        )
    app.echo_info(f"Import workspace from {file} ... ", nl=False)
    import_workspace(file, marshalling_plugin)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="reload")
@click.pass_obj
def reload_command(app):
    """Reload the workspace from the backend."""
    ctx = click.get_current_context()
    if ctx.auto_envvar_prefix == "CMEMC_WORKSPACE_RELOAD":
        app.echo_warning(
            "The command workspace reload is deprecated. "
            "Please use 'cmemc admin workspace reload' instead."
        )
    app.echo_info("Reload workspace  ... ", nl=False)
    reload_workspace()
    app.echo_success("done")


@click.group(cls=CmemcGroup, )
def workspace():
    """Import, export and reload the project workspace."""


workspace.add_command(export_command)
workspace.add_command(import_command)
workspace.add_command(reload_command)
workspace.add_command(python)
